"""
The Model class and related functionality.
"""
from abc import ABCMeta
import warnings

import tensorflow as tf

from cleverhans import utils_tf


class Model(object):
  """
  An abstract interface for model wrappers that exposes model symbols
  needed for making an attack. This abstraction removes the dependency on
  any specific neural network package (e.g. Keras) from the core
  code of CleverHans. It can also simplify exposing the hidden features of a
  model when a specific package does not directly expose them.
  """
  __metaclass__ = ABCMeta
  O_LOGITS, O_PROBS, O_FEATURES = 'logits probs features'.split()

  def __init__(self, scope=None, nb_classes=None, hparams=None,
               needs_dummy_fprop=False):
    """
    Constructor.
    :param scope: str, the name of model.
    :param nb_classes: integer, the number of classes.
    :param hparams: dict, hyper-parameters for the model.
    :needs_dummy_fprop: bool, if True the model's parameters are not
        created until fprop is called.
    """
    self.scope = scope or self.__class__.__name__
    self.nb_classes = nb_classes
    self.hparams = hparams or {}
    self.needs_dummy_fprop = needs_dummy_fprop

  def __call__(self, *args, **kwargs):
    """
    For compatibility with functions used as model definitions (taking
    an input tensor and returning the tensor giving the output
    of the model on that input).
    """

    warnings.warn("Model.__call__ is deprecated. "
                  "The call is ambiguous as to whether the output should "
                  "be logits or probabilities, and getting the wrong one "
                  "can cause serious problems. "
                  "The output actually is probabilities, which are a very "
                  "dangerous thing to use as part of any interface for "
                  "cleverhans, because softmax probabilities are prone "
                  "to gradient masking."
                  "On or after 2019-04-24, this method will change to raise "
                  "an exception explaining why Model.__call__ should not be "
                  "used.")

    return self.get_probs(*args, **kwargs)

  def get_logits(self, x, **kwargs):
    """
    :param x: A symbolic representation (Tensor) of the network input
    :return: A symbolic representation (Tensor) of the output logits
    (i.e., the values fed as inputs to the softmax layer).
    """
    outputs = self.fprop(x, **kwargs)
    if self.O_LOGITS in outputs:
      return outputs[self.O_LOGITS]
    raise NotImplementedError(str(type(self)) + "must implement `get_logits`"
                              " or must define a " + self.O_LOGITS +
                              " output in `fprop`")

  def get_predicted_class(self, x, **kwargs):
    """
    :param x: A symbolic representation (Tensor) of the network input
    :return: A symbolic representation (Tensor) of the predicted label
    """
    return tf.argmax(self.get_logits(x, **kwargs), axis=1)

  def get_probs(self, x, **kwargs):
    """
    :param x: A symbolic representation (Tensor) of the network input
    :return: A symbolic representation (Tensor) of the output
    probabilities (i.e., the output values produced by the softmax layer).
    """
    d = self.fprop(x, **kwargs)
    if self.O_PROBS in d:
      output = d[self.O_PROBS]
      min_prob = tf.reduce_min(output)
      max_prob = tf.reduce_max(output)
      asserts = [utils_tf.assert_greater_equal(min_prob,
                                               tf.cast(0., min_prob.dtype)),
                 utils_tf.assert_less_equal(max_prob,
                                            tf.cast(1., min_prob.dtype))]
      with tf.control_dependencies(asserts):
        output = tf.identity(output)
      return output
    elif self.O_LOGITS in d:
      return tf.nn.softmax(logits=d[self.O_LOGITS])
    else:
      raise ValueError('Cannot find probs or logits.')

  def fprop(self, x, **kwargs):
    """
    Forward propagation to compute the model outputs.
    :param x: A symbolic representation of the network input
    :return: A dictionary mapping layer names to the symbolic
             representation of their output.
    """
    raise NotImplementedError('`fprop` not implemented.')

  def get_params(self):
    """
    Provides access to the model's parameters.
    :return: A list of all Variables defining the model parameters.
    """

    if hasattr(self, 'params'):
      return list(self.params)

    # Catch eager execution and assert function overload.
    try:
      if tf.executing_eagerly():
        raise NotImplementedError("For Eager execution - get_params "
                                  "must be overridden.")
    except AttributeError:
      pass

    # For graph-based execution
    scope_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES,
                                   self.scope + "/")

    if len(scope_vars) == 0:
      self.make_params()
      scope_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES,
                                     self.scope + "/")
      assert len(scope_vars) > 0

    # Make sure no parameters have been added or removed
    if hasattr(self, "num_params"):
      if self.num_params != len(scope_vars):
        print("Scope: ", self.scope)
        print("Expected " + str(self.num_params) + " variables")
        print("Got " + str(len(scope_vars)))
        for var in scope_vars:
          print("\t" + str(var))
        assert False
    else:
      self.num_params = len(scope_vars)

    return scope_vars

  def make_params(self):
    """
    Create all Variables to be returned later by get_params.
    By default this is a no-op.
    Models that need their fprop to be called for their params to be
    created can set `needs_dummy_fprop=True` in the constructor.
    """

    if self.needs_dummy_fprop:
      if hasattr(self, "_dummy_input"):
        return
      self._dummy_input = self.make_input_placeholder()
      self.fprop(self._dummy_input)

  def get_layer_names(self):
    """Return the list of exposed layers for this model."""
    raise NotImplementedError

  def get_layer(self, x, layer, **kwargs):
    """Return a layer output.
    :param x: tensor, the input to the network.
    :param layer: str, the name of the layer to compute.
    :param **kwargs: dict, extra optional params to pass to self.fprop.
    :return: the content of layer `layer`
    """
    return self.fprop(x, **kwargs)[layer]

  def make_input_placeholder(self):
    """Create and return a placeholder representing an input to the model.

    This method should respect context managers (e.g. "with tf.device")
    and should not just return a reference to a single pre-created
    placeholder.
    """

    raise NotImplementedError(str(type(self)) + " does not implement "
                              "make_input_placeholder")

  def make_label_placeholder(self):
    """Create and return a placeholder representing class labels.

    This method should respect context managers (e.g. "with tf.device")
    and should not just return a reference to a single pre-created
    placeholder.
    """

    raise NotImplementedError(str(type(self)) + " does not implement "
                              "make_label_placeholder")

  def __hash__(self):
    return hash(id(self))

  def __eq__(self, other):
    return self is other


class CallableModelWrapper(Model):
  """A wrapper that turns a callable into a valid Model"""

  def __init__(self, callable_fn, output_layer):
    """
    Wrap a callable function that takes a tensor as input and returns
    a tensor as output with the given layer name.
    :param callable_fn: The callable function taking a tensor and
                        returning a given layer as output.
    :param output_layer: A string of the output layer returned by the
                         function. (Usually either "probs" or "logits".)
    """

    super(CallableModelWrapper, self).__init__()
    self.output_layer = output_layer
    self.callable_fn = callable_fn

  def fprop(self, x, **kwargs):
    output = self.callable_fn(x, **kwargs)

    # Do some sanity checking to reduce the chance that probs are used
    # as logits accidentally or vice versa
    if self.output_layer == 'probs':
      assert output.op.type == "Softmax"
      min_prob = tf.reduce_min(output)
      max_prob = tf.reduce_max(output)
      asserts = [utils_tf.assert_greater_equal(min_prob,
                                               tf.cast(0., min_prob.dtype)),
                 utils_tf.assert_less_equal(max_prob,
                                            tf.cast(1., max_prob.dtype))]
      with tf.control_dependencies(asserts):
        output = tf.identity(output)
    elif self.output_layer == 'logits':
      assert output.op.type != 'Softmax'

    return {self.output_layer: output}

def wrapper_warning():
  """
  Issue a deprecation warning. Used in multiple places that implemented
  attacks by automatically wrapping a user-supplied callable with a
  CallableModelWrapper with output_layer="probs".
  Using "probs" as any part of the attack interface is dangerous.
  We can't just change output_layer to logits because:
  - that would be a silent interface change. We'd have no way of detecting
    code that still means to use probs. Note that we can't just check whether
    the final output op is a softmax---for example, Inception puts a reshape
    after the softmax.
  - automatically wrapping user-supplied callables with output_layer='logits'
    is even worse, see `wrapper_warning_logits`
  Note: this function will be removed at the same time as the code that
  calls it.
  """
  warnings.warn("Passing a callable is deprecated, because using"
                " probabilities is dangerous. It has a high risk "
                " of causing gradient masking due to loss of precision "
                " in the softmax op. Passing a callable rather than a "
                " Model subclass will become an error on or after "
                " 2019-04-24.")

def wrapper_warning_logits():
  """
  Issue a deprecation warning. Used in multiple places that implemented
  attacks by automatically wrapping a user-supplied callable with a
  CallableModelWrapper with output_layer="logits".
  This is dangerous because it is under-the-hood automagic that the user
  may not realize has been invoked for them. If they pass a callable
  that actually outputs probs, the probs will be treated as logits,
  resulting in an incorrect cross-entropy loss and severe gradient
  masking.
  """
  warnings.warn("Passing a callable is deprecated, because it runs the "
                "risk of accidentally using probabilities in the place "
                "of logits. Please switch to passing a Model subclass "
                "so that you clearly specify which values are the logits. "
                "Passing a callable rather than a Model subclass will become "
                "an error on or after 2019-04-24.")


class NoSuchLayerError(ValueError):

  """Raised when a layer that does not exist is requested."""



"""
Functionality for evaluating expressions across entire datasets.
Includes multi-GPU support for fast evaluation.
"""

from distutils.version import LooseVersion
import warnings
import numpy as np
from six.moves import range
import tensorflow as tf

import cleverhans
from cleverhans import canary
from cleverhans.utils import create_logger
from cleverhans.utils_tf import infer_devices


def accuracy(sess, model, x, y, batch_size=None, devices=None, feed=None,
             attack=None, attack_params=None):
  """
  Compute the accuracy of a TF model on some data
  :param sess: TF session to use when training the graph
  :param model: cleverhans.model.Model instance
  :param x: numpy array containing input examples (e.g. MNIST().x_test )
  :param y: numpy array containing example labels (e.g. MNIST().y_test )
  :param batch_size: Number of examples to use in a single evaluation batch.
      If not specified, this function will use a reasonable guess and
      may run out of memory.
      When choosing the batch size, keep in mind that the batch will
      be divided up evenly among available devices. If you can fit 128
      examples in memory on one GPU and you have 8 GPUs, you probably
      want to use a batch size of 1024 (unless a different batch size
      runs faster with the ops you are using, etc.)
  :param devices: An optional list of string device names to use.
    If not specified, this function will use all visible GPUs.
  :param feed: An optional dictionary that is appended to the feeding
           dictionary before the session runs. Can be used to feed
           the learning phase of a Keras model for instance.
  :param attack: cleverhans.attack.Attack
    Optional. If no attack specified, evaluates the model on clean data.
    If attack is specified, evaluates the model on adversarial examples
    created by the attack.
  :param attack_params: dictionary
    If attack is specified, this dictionary is passed to attack.generate
    as keyword arguments.
  :return: a float with the accuracy value
  """

  _check_x(x)
  _check_y(y)
  if x.shape[0] != y.shape[0]:
    raise ValueError("Number of input examples and labels do not match.")

  factory = _CorrectFactory(model, attack, attack_params)

  correct, = batch_eval_multi_worker(sess, factory, [x, y],
                                     batch_size=batch_size, devices=devices,
                                     feed=feed)

  return correct.mean()


def correctness_and_confidence(sess, model, x, y, batch_size=None,
                               devices=None, feed=None, attack=None,
                               attack_params=None):
  """
  Report whether the model is correct and its confidence on each example in
  a dataset.
  :param sess: tf.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing input examples (e.g. MNIST().x_test )
  :param y: numpy array containing example labels (e.g. MNIST().y_test )
  :param batch_size: Number of examples to use in a single evaluation batch.
      If not specified, this function will use a reasonable guess and
      may run out of memory.
      When choosing the batch size, keep in mind that the batch will
      be divided up evenly among available devices. If you can fit 128
      examples in memory on one GPU and you have 8 GPUs, you probably
      want to use a batch size of 1024 (unless a different batch size
      runs faster with the ops you are using, etc.)
  :param devices: An optional list of string device names to use.
    If not specified, this function will use all visible GPUs.
  :param feed: An optional dictionary that is appended to the feeding
           dictionary before the session runs. Can be used to feed
           the learning phase of a Keras model for instance.
  :param attack: cleverhans.attack.Attack
    Optional. If no attack specified, evaluates the model on clean data.
    If attack is specified, evaluates the model on adversarial examples
    created by the attack.
  :param attack_params: dictionary
    If attack is specified, this dictionary is passed to attack.generate
    as keyword arguments.
  :return:
    an ndarray of bools indicating whether each example is correct
    an ndarray of probabilities assigned to the prediction for each example
  """

  _check_x(x)
  _check_y(y)
  if x.shape[0] != y.shape[0]:
    raise ValueError("Number of input examples and labels do not match.")

  factory = _CorrectAndProbFactory(model, attack, attack_params)

  out = batch_eval_multi_worker(sess, factory, [x, y], batch_size=batch_size,
                                devices=devices, feed=feed)

  correctness, confidence = out

  assert correctness.shape == (x.shape[0],)
  assert confidence.shape == (x.shape[0],)
  min_confidence = confidence.min()
  if min_confidence < 0.:
    raise ValueError("Model does not return valid probabilities: " +
                     str(min_confidence))
  max_confidence = confidence.max()
  if max_confidence > 1.:
    raise ValueError("Model does not return valid probablities: " +
                     str(max_confidence))
  assert confidence.min() >= 0., confidence.min()

  return out

def run_attack(sess, model, x, y, attack, attack_params, batch_size=None,
               devices=None, feed=None, pass_y=False):
  """
  Run attack on every example in a dataset.
  :param sess: tf.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing input examples (e.g. MNIST().x_test )
  :param y: numpy array containing example labels (e.g. MNIST().y_test )
  :param attack: cleverhans.attack.Attack
  :param attack_params: dictionary
    passed to attack.generate as keyword arguments.
  :param batch_size: Number of examples to use in a single evaluation batch.
      If not specified, this function will use a reasonable guess and
      may run out of memory.
      When choosing the batch size, keep in mind that the batch will
      be divided up evenly among available devices. If you can fit 128
      examples in memory on one GPU and you have 8 GPUs, you probably
      want to use a batch size of 1024 (unless a different batch size
      runs faster with the ops you are using, etc.)
  :param devices: An optional list of string device names to use.
    If not specified, this function will use all visible GPUs.
  :param feed: An optional dictionary that is appended to the feeding
           dictionary before the session runs. Can be used to feed
           the learning phase of a Keras model for instance.
  :param pass_y: bool. If true pass 'y' to `attack.generate`
  :return:
    an ndarray of bools indicating whether each example is correct
    an ndarray of probabilities assigned to the prediction for each example
  """

  _check_x(x)
  _check_y(y)

  factory = _AttackFactory(model, attack, attack_params, pass_y)

  out, = batch_eval_multi_worker(sess, factory, [x, y], batch_size=batch_size,
                                 devices=devices, feed=feed)
  return out


def batch_eval_multi_worker(sess, graph_factory, numpy_inputs, batch_size=None,
                            devices=None, feed=None):
  """
  Generic computation engine for evaluating an expression across a whole
  dataset, divided into batches.

  This function assumes that the work can be parallelized with one worker
  device handling one batch of data. If you need multiple devices per
  batch, use `batch_eval`.

  The tensorflow graph for multiple workers is large, so the first few
  runs of the graph will be very slow. If you expect to run the graph
  few times (few calls to `batch_eval_multi_worker` that each run few
  batches) the startup cost might dominate the runtime, and it might be
  preferable to use the single worker `batch_eval` just because its
  startup cost will be lower.

  :param sess: tensorflow Session
  :param graph_factory: callable
      When called, returns (tf_inputs, tf_outputs) where:
          tf_inputs is a list of placeholders to feed from the dataset
          tf_outputs is a list of tf tensors to calculate
      Example: tf_inputs is [x, y] placeholders, tf_outputs is [accuracy].
      This factory must make new tensors when called, rather than, e.g.
      handing out a reference to existing tensors.
      This factory must make exactly equivalent expressions every time
      it is called, otherwise the results of `batch_eval` will vary
      depending on how work is distributed to devices.
      This factory must respect "with tf.device()" context managers
      that are active when it is called, otherwise work will not be
      distributed to devices correctly.
  :param numpy_inputs:
      A list of numpy arrays defining the dataset to be evaluated.
      The list should have the same length as tf_inputs.
      Each array should have the same number of examples (shape[0]).
      Example: numpy_inputs is [MNIST().x_test, MNIST().y_test]
  :param batch_size: Number of examples to use in a single evaluation batch.
      If not specified, this function will use a reasonable guess and
      may run out of memory.
      When choosing the batch size, keep in mind that the batch will
      be divided up evenly among available devices. If you can fit 128
      examples in memory on one GPU and you have 8 GPUs, you probably
      want to use a batch size of 1024 (unless a different batch size
      runs faster with the ops you are using, etc.)
  :param devices: List of devices to run on. If unspecified, uses all
      available GPUs if any GPUS are available, otherwise uses CPUs.
  :param feed: An optional dictionary that is appended to the feeding
           dictionary before the session runs. Can be used to feed
           the learning phase of a Keras model for instance.
  :returns: List of numpy arrays corresponding to the outputs produced by
      the graph_factory
  """
  canary.run_canary()
  global _batch_eval_multi_worker_cache

  devices = infer_devices(devices)

  if batch_size is None:
    # For big models this might result in OOM and then the user
    # should just specify batch_size
    batch_size = len(devices) * DEFAULT_EXAMPLES_PER_DEVICE

  n = len(numpy_inputs)
  assert n > 0
  m = numpy_inputs[0].shape[0]
  for i in range(1, n):
    m_i = numpy_inputs[i].shape[0]
    if m != m_i:
      raise ValueError("All of numpy_inputs must have the same number of examples, but the first one has " + str(m)
                       + " examples and input " + str(i) + " has " + str(m_i) + "examples.")
  out = []

  replicated_tf_inputs = []
  replicated_tf_outputs = []
  p = None

  num_devices = len(devices)
  assert batch_size % num_devices == 0
  device_batch_size = batch_size // num_devices

  cache_key = (graph_factory, tuple(devices))
  if cache_key in _batch_eval_multi_worker_cache:
    # Retrieve graph for multi-GPU inference from cache.
    # This avoids adding tf ops to the graph
    packed = _batch_eval_multi_worker_cache[cache_key]
    replicated_tf_inputs, replicated_tf_outputs = packed
    p = len(replicated_tf_outputs[0])
    assert p > 0
  else:
    # This graph has not been built before.
    # Build it now.

    for device in devices:
      with tf.device(device):
        tf_inputs, tf_outputs = graph_factory()
        assert len(tf_inputs) == n
        if p is None:
          p = len(tf_outputs)
          assert p > 0
        else:
          assert len(tf_outputs) == p
        replicated_tf_inputs.append(tf_inputs)
        replicated_tf_outputs.append(tf_outputs)
    del tf_inputs
    del tf_outputs
    # Store the result in the cache
    packed = replicated_tf_inputs, replicated_tf_outputs
    _batch_eval_multi_worker_cache[cache_key] = packed
  for _ in range(p):
    out.append([])
  flat_tf_outputs = []
  for output in range(p):
    for dev_idx in range(num_devices):
      flat_tf_outputs.append(replicated_tf_outputs[dev_idx][output])

  # pad data to have # examples be multiple of batch size
  # we discard the excess later
  num_batches = int(np.ceil(float(m) / batch_size))
  needed_m = num_batches * batch_size
  excess = needed_m - m
  if excess > m:
    raise NotImplementedError(("Your batch size (%(batch_size)d) is bigger"
                               " than the dataset (%(m)d), this function is "
                               "probably overkill.") % locals())

  def pad(array):
    """Pads an array with replicated examples to have `excess` more entries"""
    if excess > 0:
      array = np.concatenate((array, array[:excess]), axis=0)
    return array
  numpy_inputs = [pad(numpy_input) for numpy_input in numpy_inputs]
  orig_m = m
  m = needed_m

  for start in range(0, m, batch_size):
    batch = start // batch_size
    if batch % 100 == 0 and batch > 0:
      _logger.debug("Batch " + str(batch))

    # Compute batch start and end indices
    end = start + batch_size
    numpy_input_batches = [numpy_input[start:end]
                           for numpy_input in numpy_inputs]
    feed_dict = {}
    for dev_idx, tf_inputs in enumerate(replicated_tf_inputs):
      for tf_input, numpy_input in zip(tf_inputs, numpy_input_batches):
        dev_start = dev_idx * device_batch_size
        dev_end = (dev_idx + 1) * device_batch_size
        value = numpy_input[dev_start:dev_end]
        assert value.shape[0] == device_batch_size
        feed_dict[tf_input] = value
    if feed is not None:
      feed_dict.update(feed)
    flat_output_batches = sess.run(flat_tf_outputs, feed_dict=feed_dict)
    for e in flat_output_batches:
      assert e.shape[0] == device_batch_size, e.shape

    output_batches = []
    for output in range(p):
      o_start = output * num_devices
      o_end = (output + 1) * num_devices
      device_values = flat_output_batches[o_start:o_end]
      assert len(device_values) == num_devices
      output_batches.append(device_values)

    for out_elem, device_values in zip(out, output_batches):
      assert len(device_values) == num_devices, (len(device_values),
                                                 num_devices)
      for device_value in device_values:
        assert device_value.shape[0] == device_batch_size
      out_elem.extend(device_values)

  out = [np.concatenate(x, axis=0) for x in out]
  for e in out:
    assert e.shape[0] == m, e.shape

  # Trim off the examples we used to pad up to batch size
  out = [e[:orig_m] for e in out]
  assert len(out) == p, (len(out), p)

  return out


def batch_eval(sess, tf_inputs, tf_outputs, numpy_inputs, batch_size=None,
               feed=None,
               args=None):
  """
  A helper function that computes a tensor on numpy inputs by batches.
  This version uses exactly the tensorflow graph constructed by the
  caller, so the caller can place specific ops on specific devices
  to implement model parallelism.
  Most users probably prefer `batch_eval_multi_worker` which maps
  a single-device expression to multiple devices in order to evaluate
  faster by parallelizing across data.

  :param sess: tf Session to use
  :param tf_inputs: list of tf Placeholders to feed from the dataset
  :param tf_outputs: list of tf tensors to calculate
  :param numpy_inputs: list of numpy arrays defining the dataset
  :param batch_size: int, batch size to use for evaluation
      If not specified, this function will try to guess the batch size,
      but might get an out of memory error or run the model with an
      unsupported batch size, etc.
  :param feed: An optional dictionary that is appended to the feeding
           dictionary before the session runs. Can be used to feed
           the learning phase of a Keras model for instance.
  :param args: dict or argparse `Namespace` object.
              Deprecated and included only for backwards compatibility.
               Should contain `batch_size`
  """

  if args is not None:
    warnings.warn("`args` is deprecated and will be removed on or "
                  "after 2019-03-09. Pass `batch_size` directly.")
    if "batch_size" in args:
      assert batch_size is None
      batch_size = args["batch_size"]

  if batch_size is None:
    batch_size = DEFAULT_EXAMPLES_PER_DEVICE

  n = len(numpy_inputs)
  assert n > 0
  assert n == len(tf_inputs)
  m = numpy_inputs[0].shape[0]
  for i in range(1, n):
    assert numpy_inputs[i].shape[0] == m
  out = []
  for _ in tf_outputs:
    out.append([])
  for start in range(0, m, batch_size):
    batch = start // batch_size
    if batch % 100 == 0 and batch > 0:
      _logger.debug("Batch " + str(batch))

    # Compute batch start and end indices
    start = batch * batch_size
    end = start + batch_size
    numpy_input_batches = [numpy_input[start:end]
                           for numpy_input in numpy_inputs]
    cur_batch_size = numpy_input_batches[0].shape[0]
    assert cur_batch_size <= batch_size
    for e in numpy_input_batches:
      assert e.shape[0] == cur_batch_size

    feed_dict = dict(zip(tf_inputs, numpy_input_batches))
    if feed is not None:
      feed_dict.update(feed)
    numpy_output_batches = sess.run(tf_outputs, feed_dict=feed_dict)
    for e in numpy_output_batches:
      assert e.shape[0] == cur_batch_size, e.shape
    for out_elem, numpy_output_batch in zip(out, numpy_output_batches):
      out_elem.append(numpy_output_batch)

  out = [np.concatenate(x, axis=0) for x in out]
  for e in out:
    assert e.shape[0] == m, e.shape
  return out


DEFAULT_EXAMPLES_PER_DEVICE = 128


class _CorrectFactory(object):
  """
  A factory for an expression for one bool per example indicating
  whether each example is correct.
  """

  def __init__(self, model, attack=None, attack_params=None):
    if attack_params is None:
      attack_params = {}
    self.model = model
    self.attack = attack
    self.attack_params = attack_params
    hashable_attack_params = tuple((key, attack_params[key]) for key
                                   in sorted(attack_params.keys()))
    self.properties_to_hash = (model, attack, hashable_attack_params)

  def __hash__(self):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    return self.properties_to_hash.__hash__()

  def __eq__(self, other):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    if not isinstance(other, _CorrectFactory):
      return False
    return self.properties_to_hash == other.properties_to_hash

  def __call__(self):
    x_batch = self.model.make_input_placeholder()
    y_batch = self.model.make_label_placeholder()

    if LooseVersion(tf.__version__) < LooseVersion('1.0.0'):
      raise NotImplementedError()

    if self.attack is None:
      x_input = x_batch
    else:
      attack_params = self.attack_params
      if attack_params is None:
        attack_params = {}
      x_input = self.attack.generate(x_batch, y=y_batch, **attack_params)

    predictions = self.model.get_probs(x_input)
    correct = tf.equal(tf.argmax(y_batch, axis=-1),
                       tf.argmax(predictions, axis=-1))

    return (x_batch, y_batch), (correct,)


class _CorrectAndProbFactory(object):
  """
  A factory for an expression for the following tuple per (optionally
  adversarial) example:
    - bool per indicating whether each the example was classified correctly
    - probability assigned to that prediction
  """

  def __init__(self, model, attack=None, attack_params=None):
    if attack_params is None:
      attack_params = {}
    self.model = model
    self.attack = attack
    self.attack_params = attack_params
    hashable_attack_params = tuple((key, attack_params[key]) for key
                                   in sorted(attack_params.keys()))
    self.properties_to_hash = (model, attack, hashable_attack_params)

  def __hash__(self):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    return self.properties_to_hash.__hash__()

  def __eq__(self, other):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    if not isinstance(other, _CorrectAndProbFactory):
      return False
    return self.properties_to_hash == other.properties_to_hash

  def __call__(self):
    x_batch = self.model.make_input_placeholder()
    y_batch = self.model.make_label_placeholder()

    if LooseVersion(tf.__version__) < LooseVersion('1.0.0'):
      raise NotImplementedError()

    if self.attack is None:
      x_input = x_batch
    else:
      attack_params = self.attack_params
      if attack_params is None:
        attack_params = {}
      x_input = self.attack.generate(x_batch, y=y_batch, **attack_params)

    predictions = self.model.get_probs(x_input)
    correct = tf.equal(tf.argmax(y_batch, axis=-1),
                       tf.argmax(predictions, axis=-1))
    max_probs = tf.reduce_max(predictions, axis=1)

    return (x_batch, y_batch), (correct, max_probs)


class _AttackFactory(object):
  """
  A factory for an expression that runs an adversarial attack

  :param model: cleverhans.model.Model
  :param attack: cleverhans.attack.Attack
  :param attack_params: dict of arguments to pass to attack.generate
  :param pass_y: bool. If True, pass y to the attack.
    (Some untargeted attacks prefer to infer y to avoid label leaking.
    Targeted attacks require that y not be passed)
  """

  def __init__(self, model, attack, attack_params=None, pass_y=False):
    assert isinstance(model, cleverhans.model.Model)
    if not isinstance(attack, cleverhans.attacks.Attack):
      raise TypeError("`attack` must be an instance of cleverhans.attacks."
                      "attack. Got %s with type %s " % (str(attack),
                                                        str(type(attack))))

    if attack_params is None:
      attack_params = {}
    self.model = model
    self.attack = attack
    self.attack_params = attack_params
    self.pass_y = pass_y
    hashable_attack_params = tuple((key, attack_params[key]) for key
                                   in sorted(attack_params.keys()))
    self.properties_to_hash = (model, attack, hashable_attack_params)

  def __hash__(self):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    return self.properties_to_hash.__hash__()

  def __eq__(self, other):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    if not isinstance(other, _AttackFactory):
      return False
    return self.properties_to_hash == other.properties_to_hash

  def __call__(self):
    x_batch = self.model.make_input_placeholder()
    y_batch = self.model.make_label_placeholder()

    attack_params = self.attack_params
    if attack_params is None:
      attack_params = {}
    if self.pass_y:
      x_adv = self.attack.generate(x_batch, y=y_batch, **attack_params)
    else:
      # Some code checks the keys of kwargs, rather than checking if
      # y is None, so we need to truly not pass y at all, rather than
      # just passing a None value for it.
      x_adv = self.attack.generate(x_batch, **attack_params)

    return (x_batch, y_batch), tuple([x_adv])

_logger = create_logger("cleverhans.evaluation")

# Cache for storing output of `batch_eval_multi_worker`'s calls to
# `graph_factory`, to avoid making the tf graph too big
_batch_eval_multi_worker_cache = {}

def _check_x(x):
  """
  Makes sure an `x` argument is a valid numpy dataset.
  """
  if not isinstance(x, np.ndarray):
    raise TypeError("x must be a numpy array. Typically x contains "
                    "the entire test set inputs.")

def _check_y(y):
  """
  Makes sure a `y` argument is a vliad numpy dataset.
  """
  if not isinstance(y, np.ndarray):
    raise TypeError("y must be numpy array. Typically y contains "
                    "the entire test set labels.")



"""
Runs multiple attacks against each example.

References: https://openreview.net/forum?id=H1g0piA9tQ
            https://arxiv.org/abs/1811.03685
"""
import copy
import logging
import time

import numpy as np
import six
from six.moves import range
import tensorflow as tf

from cleverhans.attacks import Noise
from cleverhans.attacks import ProjectedGradientDescent
from cleverhans.attacks import SPSA
from cleverhans.evaluation import correctness_and_confidence
from cleverhans.evaluation import batch_eval_multi_worker, run_attack
from cleverhans.model import Model
from cleverhans import serial
from cleverhans.utils import create_logger, deep_copy, safe_zip
from cleverhans.utils_tf import infer_devices
from cleverhans.confidence_report import ConfidenceReport
from cleverhans.confidence_report import ConfidenceReportEntry
from cleverhans.confidence_report import print_stats

_logger = create_logger("train")
_logger.setLevel(logging.INFO)

devices = infer_devices()
num_devices = len(devices)
DEFAULT_EXAMPLES_PER_DEVICE = 128
BATCH_SIZE = DEFAULT_EXAMPLES_PER_DEVICE * num_devices
REPORT_TIME_INTERVAL = 60

# TODO: lower priority: make it possible to initialize one attack with
# the output of an earlier attack


def single_run_max_confidence_recipe(sess, model, x, y, nb_classes, eps,
                                     clip_min, clip_max, eps_iter, nb_iter,
                                     report_path,
                                     batch_size=BATCH_SIZE,
                                     eps_iter_small=None):
  """A reasonable attack bundling recipe for a max norm threat model and
  a defender that uses confidence thresholding. This recipe uses both
  uniform noise and randomly-initialized PGD targeted attacks.

  References:
  https://openreview.net/forum?id=H1g0piA9tQ

  This version runs each attack (noise, targeted PGD for each class with
  nb_iter iterations, target PGD for each class with 25X more iterations)
  just once and then stops. See `basic_max_confidence_recipe` for a version
  that runs indefinitely.

  :param sess: tf.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param nb_classes: int, number of classes
  :param eps: float, maximum size of perturbation (measured by max norm)
  :param eps_iter: float, step size for one version of PGD attacks
    (will also run another version with eps_iter_small step size)
  :param nb_iter: int, number of iterations for the cheaper PGD attacks
    (will also run another version with 25X more iterations)
  :param report_path: str, the path that the report will be saved to.
  :param batch_size: int, the total number of examples to run simultaneously
  :param eps_iter_small: optional, float.
    The second version of the PGD attack is run with 25 * nb_iter iterations
    and eps_iter_small step size. If eps_iter_small is not specified it is
    set to eps_iter / 25.
  """
  noise_attack = Noise(model, sess)
  pgd_attack = ProjectedGradientDescent(model, sess)
  threat_params = {"eps": eps, "clip_min": clip_min, "clip_max": clip_max}
  noise_attack_config = AttackConfig(noise_attack, threat_params, "noise")
  attack_configs = [noise_attack_config]
  pgd_attack_configs = []
  pgd_params = copy.copy(threat_params)
  pgd_params["eps_iter"] = eps_iter
  pgd_params["nb_iter"] = nb_iter
  assert batch_size % num_devices == 0
  dev_batch_size = batch_size // num_devices
  ones = tf.ones(dev_batch_size, tf.int32)
  expensive_pgd = []
  if eps_iter_small is None:
    eps_iter_small = eps_iter / 25.
  for cls in range(nb_classes):
    cls_params = copy.copy(pgd_params)
    cls_params['y_target'] = tf.to_float(tf.one_hot(ones * cls, nb_classes))
    cls_attack_config = AttackConfig(pgd_attack, cls_params, "pgd_" + str(cls))
    pgd_attack_configs.append(cls_attack_config)
    expensive_params = copy.copy(cls_params)
    expensive_params["eps_iter"] = eps_iter_small
    expensive_params["nb_iter"] *= 25.
    expensive_config = AttackConfig(
        pgd_attack, expensive_params, "expensive_pgd_" + str(cls))
    expensive_pgd.append(expensive_config)
  attack_configs = [noise_attack_config] + pgd_attack_configs + expensive_pgd
  new_work_goal = {config: 1 for config in attack_configs}
  goals = [MaxConfidence(t=1., new_work_goal=new_work_goal)]
  bundle_attacks(sess, model, x, y, attack_configs, goals, report_path)


def basic_max_confidence_recipe(sess, model, x, y, nb_classes, eps,
                                clip_min, clip_max, eps_iter, nb_iter,
                                report_path,
                                batch_size=BATCH_SIZE,
                                eps_iter_small=None):
  """A reasonable attack bundling recipe for a max norm threat model and
  a defender that uses confidence thresholding.

  References:
  https://openreview.net/forum?id=H1g0piA9tQ

  This version runs indefinitely, updating the report on disk continuously.

  :param sess: tf.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param nb_classes: int, number of classes
  :param eps: float, maximum size of perturbation (measured by max norm)
  :param eps_iter: float, step size for one version of PGD attacks
    (will also run another version with eps_iter_small)
  :param nb_iter: int, number of iterations for one version of PGD attacks
    (will also run another version with 25X more iterations)
  :param report_path: str, the path that the report will be saved to.
  :batch_size: int, the total number of examples to run simultaneously
  :param eps_iter_small: optional, float.
    The second version of the PGD attack is run with 25 * nb_iter iterations
    and eps_iter_small step size. If eps_iter_small is not specified it is
    set to eps_iter / 25.
  """
  noise_attack = Noise(model, sess)
  pgd_attack = ProjectedGradientDescent(model, sess)
  threat_params = {"eps": eps, "clip_min": clip_min, "clip_max": clip_max}
  noise_attack_config = AttackConfig(noise_attack, threat_params)
  attack_configs = [noise_attack_config]
  pgd_attack_configs = []
  pgd_params = copy.copy(threat_params)
  pgd_params["eps_iter"] = eps_iter
  pgd_params["nb_iter"] = nb_iter
  assert batch_size % num_devices == 0
  dev_batch_size = batch_size // num_devices
  ones = tf.ones(dev_batch_size, tf.int32)
  expensive_pgd = []
  if eps_iter_small is None:
    eps_iter_small = eps_iter / 25.
  for cls in range(nb_classes):
    cls_params = copy.copy(pgd_params)
    cls_params['y_target'] = tf.to_float(tf.one_hot(ones * cls, nb_classes))
    cls_attack_config = AttackConfig(pgd_attack, cls_params, "pgd_" + str(cls))
    pgd_attack_configs.append(cls_attack_config)
    expensive_params = copy.copy(cls_params)
    expensive_params["eps_iter"] = eps_iter_small
    expensive_params["nb_iter"] *= 25.
    expensive_config = AttackConfig(
        pgd_attack, expensive_params, "expensive_pgd_" + str(cls))
    expensive_pgd.append(expensive_config)
  attack_configs = [noise_attack_config] + pgd_attack_configs + expensive_pgd
  new_work_goal = {config: 5 for config in attack_configs}
  pgd_work_goal = {config: 5 for config in pgd_attack_configs}
  goals = [Misclassify(new_work_goal={noise_attack_config: 50}),
           Misclassify(new_work_goal=pgd_work_goal),
           MaxConfidence(t=0.5, new_work_goal=new_work_goal),
           MaxConfidence(t=0.75, new_work_goal=new_work_goal),
           MaxConfidence(t=0.875, new_work_goal=new_work_goal),
           MaxConfidence(t=0.9375, new_work_goal=new_work_goal),
           MaxConfidence(t=0.96875, new_work_goal=new_work_goal),
           MaxConfidence(t=0.984375, new_work_goal=new_work_goal),
           MaxConfidence(t=1.)]
  bundle_attacks(sess, model, x, y, attack_configs, goals, report_path)
  # This runs forever


def fixed_max_confidence_recipe(sess, model, x, y, nb_classes, eps,
                                clip_min, clip_max, eps_iter, nb_iter,
                                report_path,
                                batch_size=BATCH_SIZE):
  """A reasonable attack bundling recipe for a max norm threat model and
  a defender that uses confidence thresholding.

  References:
  https://openreview.net/forum?id=H1g0piA9tQ

  This version runs each attack a fixed number of times.
  It is more exhaustive than `single_run_max_confidence_recipe` but because
  it uses a fixed budget rather than running indefinitely it is more
  appropriate for making fair comparisons between two models.

  :param sess: tf.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param nb_classes: int, number of classes
  :param eps: float, maximum size of perturbation (measured by max norm)
  :param eps_iter: float, step size for one version of PGD attacks
    (will also run another version with 25X smaller step size)
  :param nb_iter: int, number of iterations for one version of PGD attacks
    (will also run another version with 25X more iterations)
  :param report_path: str, the path that the report will be saved to.
  :batch_size: int, the total number of examples to run simultaneously
  """
  noise_attack = Noise(model, sess)
  pgd_attack = ProjectedGradientDescent(model, sess)
  threat_params = {"eps": eps, "clip_min": clip_min, "clip_max": clip_max}
  noise_attack_config = AttackConfig(noise_attack, threat_params)
  attack_configs = [noise_attack_config]
  pgd_attack_configs = []
  pgd_params = copy.copy(threat_params)
  pgd_params["eps_iter"] = eps_iter
  pgd_params["nb_iter"] = nb_iter
  assert batch_size % num_devices == 0
  dev_batch_size = batch_size // num_devices
  ones = tf.ones(dev_batch_size, tf.int32)
  expensive_pgd = []
  for cls in range(nb_classes):
    cls_params = copy.copy(pgd_params)
    cls_params['y_target'] = tf.to_float(tf.one_hot(ones * cls, nb_classes))
    cls_attack_config = AttackConfig(pgd_attack, cls_params, "pgd_" + str(cls))
    pgd_attack_configs.append(cls_attack_config)
    expensive_params = copy.copy(cls_params)
    expensive_params["eps_iter"] /= 25.
    expensive_params["nb_iter"] *= 25.
    expensive_config = AttackConfig(
        pgd_attack, expensive_params, "expensive_pgd_" + str(cls))
    expensive_pgd.append(expensive_config)
  attack_configs = [noise_attack_config] + pgd_attack_configs + expensive_pgd
  new_work_goal = {config: 5 for config in attack_configs}
  pgd_work_goal = {config: 5 for config in pgd_attack_configs}
  # TODO: lower priority: make sure bundler won't waste time running targeted
  # attacks on examples where the target class is the true class.
  goals = [Misclassify(new_work_goal={noise_attack_config: 50}),
           Misclassify(new_work_goal=pgd_work_goal),
           MaxConfidence(t=0.5, new_work_goal=new_work_goal),
           MaxConfidence(t=0.75, new_work_goal=new_work_goal),
           MaxConfidence(t=0.875, new_work_goal=new_work_goal),
           MaxConfidence(t=0.9375, new_work_goal=new_work_goal),
           MaxConfidence(t=0.96875, new_work_goal=new_work_goal),
           MaxConfidence(t=0.984375, new_work_goal=new_work_goal),
           MaxConfidence(t=1., new_work_goal=new_work_goal)]
  bundle_attacks(sess, model, x, y, attack_configs, goals, report_path)


def random_search_max_confidence_recipe(sess, model, x, y, eps,
                                        clip_min, clip_max,
                                        report_path, batch_size=BATCH_SIZE,
                                        num_noise_points=10000):
  """Max confidence using random search.

  References:
  https://openreview.net/forum?id=H1g0piA9tQ
    Describes the max_confidence procedure used for the bundling in this recipe
  https://arxiv.org/abs/1802.00420
    Describes using random search with 1e5 or more random points to avoid
    gradient masking.

  :param sess: tf.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param nb_classes: int, number of classes
  :param eps: float, maximum size of perturbation (measured by max norm)
  :param eps_iter: float, step size for one version of PGD attacks
    (will also run another version with 25X smaller step size)
  :param nb_iter: int, number of iterations for one version of PGD attacks
    (will also run another version with 25X more iterations)
  :param report_path: str, the path that the report will be saved to.
  :batch_size: int, the total number of examples to run simultaneously
  """
  noise_attack = Noise(model, sess)
  threat_params = {"eps": eps, "clip_min": clip_min, "clip_max": clip_max}
  noise_attack_config = AttackConfig(noise_attack, threat_params)
  attack_configs = [noise_attack_config]
  assert batch_size % num_devices == 0
  new_work_goal = {noise_attack_config: num_noise_points}
  goals = [MaxConfidence(t=1., new_work_goal=new_work_goal)]
  bundle_attacks(sess, model, x, y, attack_configs, goals, report_path)


class AttackConfig(object):
  """
  An attack and associated parameters.
  :param attack: cleverhans.attacks.Attack
  :param params: dict of keyword arguments to pass to attack.generate
  :param name: str, name to be returned by __str__ / __repr__
  :param pass_y: bool, whether to pass y to `attack.generate`
  """

  def __init__(self, attack, params=None, name=None, pass_y=False):
    self.attack = attack
    self.params = params
    self.name = name
    if params is not None:
      assert isinstance(params, dict)
      for key in params:
        assert isinstance(key, six.string_types), type(key)
    self.pass_y = pass_y

  def __str__(self):
    if self.name is not None:
      return self.name
    return "AttackConfig(" + str(self.attack) + ", " + str(self.params) + ")"

  def __repr__(self):
    return self.__str__()


def bundle_attacks(sess, model, x, y, attack_configs, goals, report_path,
                   attack_batch_size=BATCH_SIZE, eval_batch_size=BATCH_SIZE):
  """
  Runs attack bundling.
  Users of cleverhans may call this function but are more likely to call
  one of the recipes above.

  Reference: https://openreview.net/forum?id=H1g0piA9tQ

  :param sess: tf.session.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param attack_configs: list of AttackConfigs to run
  :param goals: list of AttackGoals to run
    The bundler works through the goals in order, until each is satisfied.
    Some goals may never be satisfied, in which case the bundler will run
    forever, updating the report on disk as it goes.
  :param report_path: str, the path the report will be saved to
  :param attack_batch_size: int, batch size for generating adversarial examples
  :param eval_batch_size: int, batch size for evaluating the model on clean / adversarial examples
  :returns:
    adv_x: The adversarial examples, in the same format as `x`
    run_counts: dict mapping each AttackConfig to a numpy array reporting
      how many times that AttackConfig was run on each example
  """
  assert isinstance(sess, tf.Session)
  assert isinstance(model, Model)
  assert all(isinstance(attack_config, AttackConfig) for attack_config
             in attack_configs)
  assert all(isinstance(goal, AttackGoal) for goal in goals)
  assert isinstance(report_path, six.string_types)
  if x.shape[0] != y.shape[0]:
    raise ValueError("Number of input examples does not match number of labels")

  # Note: no need to precompile attacks, correctness_and_confidence
  # caches them

  run_counts = {}
  for attack_config in attack_configs:
    run_counts[attack_config] = np.zeros(x.shape[0], dtype=np.int64)

  # TODO: make an interface to pass this in if it has already been computed
  # elsewhere
  _logger.info("Running on clean data to initialize the report...")
  packed = correctness_and_confidence(sess, model, x, y, batch_size=eval_batch_size,
                                      devices=devices)
  _logger.info("...done")
  correctness, confidence = packed
  _logger.info("Accuracy: " + str(correctness.mean()))
  report = ConfidenceReport()
  report['clean'] = ConfidenceReportEntry(correctness, confidence)

  adv_x = x.copy()

  for goal in goals:
    bundle_attacks_with_goal(sess, model, x, y, adv_x, attack_configs,
                             run_counts,
                             goal, report, report_path,
                             attack_batch_size=attack_batch_size, eval_batch_size=eval_batch_size)

  # Many users will set `goals` to make this run forever, so the return
  # statement is not the primary way to get information out.
  return adv_x, run_counts

def bundle_attacks_with_goal(sess, model, x, y, adv_x, attack_configs,
                             run_counts,
                             goal, report, report_path,
                             attack_batch_size=BATCH_SIZE, eval_batch_size=BATCH_SIZE):
  """
  Runs attack bundling, working on one specific AttackGoal.
  This function is mostly intended to be called by `bundle_attacks`.

  Reference: https://openreview.net/forum?id=H1g0piA9tQ

  :param sess: tf.session.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param adv_x: numpy array containing the adversarial examples made so far
    by earlier work in the bundling process
  :param attack_configs: list of AttackConfigs to run
  :param run_counts: dict mapping AttackConfigs to numpy arrays specifying
    how many times they have been run on each example
  :param goal: AttackGoal to run
  :param report: ConfidenceReport
  :param report_path: str, the path the report will be saved to
  :param attack_batch_size: int, batch size for generating adversarial examples
  :param eval_batch_size: int, batch size for evaluating the model on adversarial examples
  """
  goal.start(run_counts)
  _logger.info("Running criteria for new goal...")
  criteria = goal.get_criteria(sess, model, adv_x, y, batch_size=eval_batch_size)
  assert 'correctness' in criteria
  _logger.info("Accuracy: " + str(criteria['correctness'].mean()))
  assert 'confidence' in criteria
  while not goal.is_satisfied(criteria, run_counts):
    run_batch_with_goal(sess, model, x, y, adv_x, criteria, attack_configs,
                        run_counts,
                        goal, report, report_path,
                        attack_batch_size=attack_batch_size)
  # Save after finishing all goals.
  # The incremental saves run on a timer. This save is needed so that the last
  # few attacks after the timer don't get discarded
  report.completed = True
  save(criteria, report, report_path, adv_x)


def run_batch_with_goal(sess, model, x, y, adv_x_val, criteria, attack_configs,
                        run_counts, goal, report, report_path,
                        attack_batch_size=BATCH_SIZE):
  """
  Runs attack bundling on one batch of data.
  This function is mostly intended to be called by
  `bundle_attacks_with_goal`.

  :param sess: tf.session.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param adv_x_val: numpy array containing the adversarial examples made so far
    by earlier work in the bundling process
  :param criteria: dict mapping string names of criteria to numpy arrays with
    their values for each example
    (Different AttackGoals track different criteria)
  :param run_counts: dict mapping AttackConfigs to numpy arrays reporting how
    many times they have been run on each example
  :param goal: the AttackGoal to work on
  :param report: dict, see `bundle_attacks_with_goal`
  :param report_path: str, path to save the report to
  """
  attack_config = goal.get_attack_config(attack_configs, run_counts, criteria)
  idxs = goal.request_examples(attack_config, criteria, run_counts,
                               attack_batch_size)
  x_batch = x[idxs]
  assert x_batch.shape[0] == attack_batch_size
  y_batch = y[idxs]
  assert y_batch.shape[0] == attack_batch_size
  adv_x_batch = run_attack(sess, model, x_batch, y_batch,
                           attack_config.attack, attack_config.params,
                           attack_batch_size, devices, pass_y=attack_config.pass_y)
  criteria_batch = goal.get_criteria(sess, model, adv_x_batch, y_batch,
                                     batch_size=min(attack_batch_size,
                                                    BATCH_SIZE))
  # This can't be parallelized because some orig examples are copied more
  # than once into the batch
  cur_run_counts = run_counts[attack_config]
  for batch_idx, orig_idx in enumerate(idxs):
    cur_run_counts[orig_idx] += 1
    should_copy = goal.new_wins(criteria, orig_idx, criteria_batch, batch_idx)
    if should_copy:
      adv_x_val[orig_idx] = adv_x_batch[batch_idx]
      for key in criteria:
        criteria[key][orig_idx] = criteria_batch[key][batch_idx]
      assert np.allclose(y[orig_idx], y_batch[batch_idx])
  report['bundled'] = ConfidenceReportEntry(criteria['correctness'], criteria['confidence'])

  should_save = False
  new_time = time.time()
  if hasattr(report, 'time'):
    if new_time - report.time > REPORT_TIME_INTERVAL:
      should_save = True
  else:
    should_save = True
  if should_save:
    report.time = new_time
    goal.print_progress(criteria, run_counts)
    save(criteria, report, report_path, adv_x_val)


def save(criteria, report, report_path, adv_x_val):
  """
  Saves the report and adversarial examples.
  :param criteria: dict, of the form returned by AttackGoal.get_criteria
  :param report: dict containing a confidence report
  :param report_path: string, filepath
  :param adv_x_val: numpy array containing dataset of adversarial examples
  """
  print_stats(criteria['correctness'], criteria['confidence'], 'bundled')

  print("Saving to " + report_path)
  serial.save(report_path, report)

  assert report_path.endswith(".joblib")
  adv_x_path = report_path[:-len(".joblib")] + "_adv.npy"
  np.save(adv_x_path, adv_x_val)


class AttackGoal(object):
  """Specifies goals for attack bundling.
  Different bundling recipes can have different priorities.
  - When choosing which examples to attack in the next batch, do we want
  to focus on examples that are not yet misclassified? Among the
  still correctly classified ones, do we want to focus on the ones that
  have not been attacked many times yet? Do we want to focus on the ones
  that have low loss / low confidence so far?
  - After an attack has been run, do we prefer the new adversarial example
  or the old one? Is the new one better if it causes higher confidence
  in the wrong prediction? If it succeeds with a smaller perturbation?
  For different use cases, the answers to these questions is different.
  Implement different AttackGoal subclasses to represent the priorities
  for your use case.
  """

  def start(self, run_counts):
    """
    Called by the bundler when it starts working on the goal.

    :param run_counts: dict mapping AttackConfigs to numpy arrays reporting
      how many times they have been run on each example.
    """
    pass

  def get_criteria(self, sess, model, advx, y, batch_size=BATCH_SIZE):
    """
    Returns a dictionary mapping the name of each criterion to a NumPy
    array containing the value of that criterion for each adversarial
    example.
    Subclasses can add extra criteria by implementing the `extra_criteria`
    method.

    :param sess: tf.session.Session
    :param model: cleverhans.model.Model
    :param adv_x: numpy array containing the adversarial examples made so far
      by earlier work in the bundling process
    :param y: numpy array containing true labels
    :param batch_size: int, batch size
    """

    names, factory = self.extra_criteria()
    factory = _CriteriaFactory(model, factory)
    results = batch_eval_multi_worker(sess, factory, [advx, y],
                                      batch_size=batch_size, devices=devices)
    names = ['correctness', 'confidence'] + names
    out = dict(safe_zip(names, results))
    return out

  def extra_criteria(self):
    """
    Subclasses implement this to specify any extra criteria they need to track.
    : returns: list of criterion names, _ExtraCriteriaFactory implementing them
    """
    return [], None

  def request_examples(self, attack_config, criteria, run_counts, batch_size):
    """
    Returns a numpy array of integer example indices to run in the next batch.
    """
    raise NotImplementedError(str(type(self)) +
                              "needs to implement request_examples")

  def is_satisfied(self, criteria, run_counts):
    """
    Returns a bool indicating whether the goal has been satisfied.
    """
    raise NotImplementedError(str(type(self)) +
                              " needs to implement is_satisfied.")

  def print_progress(self, criteria, run_counts):
    """
    Prints a progress message about how much has been done toward the goal.
    :param criteria: dict, of the format returned by get_criteria
    :param run_counts: dict mapping each AttackConfig to a numpy array
      specifying how many times it has been run for each example
    """
    print("Working on a " + self.__class__.__name__ + " goal.")

  def get_attack_config(self, attack_configs, run_counts, criteria):
    """
    Returns an AttackConfig to run on the next batch.
    """
    raise NotImplementedError(str(type(self)) +
                              " needs to implement get_attack_config")

  def new_wins(self, orig_criteria, orig_idx, new_criteria, new_idx):
    """
    Returns a bool indicating whether a new adversarial example is better
    than the pre-existing one for the same clean example.
    :param orig_criteria: dict mapping names of criteria to their value
      for each example in the whole dataset
    :param orig_idx: The position of the pre-existing example within the
      whole dataset.
    :param new_criteria: dict, like orig_criteria, but with values only
      on the latest batch of adversarial examples
    :param new_idx: The position of the new adversarial example within
      the batch
    """
    raise NotImplementedError(str(type(self))
                              + " needs to implement new_wins.")


class Misclassify(AttackGoal):
  """An AttackGoal that prioritizes misclassifying all examples.

  Times out when each attack has been run the requested number of times.
  Some examples may be attacked more than the goal number because we
  always run a full batch of attacks and sometimes the batch size is
  larger than the number of examples with attacks left to do.
  :param new_work_goal: dict
    Maps AttackConfigs to ints specifying how many times they should be
    run before timing out.
    If this dict is not specified, all attacks will be run, repeatedly,
    until all examples are misclassified (or forever if some cannot
    be changed into misclassified examples).
    If this dict is specfied, only attacks in the dict will be run.
  :param break_ties: string name of tie-breaking scheme for `new_wins`
    When two examples are misclassified, how do we choose between them?
    Currently the only scheme is 'wrong_confidence', where we prefer the
    one with higher confidence assigned to a single wrong class.
    In the future we may support other schemes like smaller perturbation
    size, higher loss, etc.
  """

  def __init__(self, new_work_goal=None, break_ties='wrong_confidence'):
    super(Misclassify, self).__init__()
    self.new_work_goal = new_work_goal
    assert all(isinstance(key, AttackConfig) for key in new_work_goal.keys())
    assert all(isinstance(value, int) for value in new_work_goal.values())
    self.rng = np.random.RandomState([2018, 10, 5, 9])
    self.break_ties = break_ties

  def start(self, run_counts):
    for key in run_counts:
      value = run_counts[key]
      assert value.ndim == 1
    _logger.info("Started working on a Misclassify goal")
    self.work_before = deep_copy(run_counts)

  def is_satisfied(self, criteria, run_counts):
    correctness = criteria['correctness']
    assert correctness.dtype == np.bool
    assert correctness.ndim == 1
    if correctness.max() == 0:
      _logger.info("Everything is misclassified! Done with Misclassify goal")
      return True
    if self.new_work_goal is None:
      return False
    correct_run_counts = self.filter(run_counts, criteria)
    correct_work_before = self.filter(self.work_before, criteria)
    unfinished = unfinished_attack_configs(self.new_work_goal,
                                           correct_work_before,
                                           correct_run_counts)
    finished = len(unfinished) == 0
    if finished:
      _logger.info("Misclassify timed out after running all requested attacks")
    else:
      pass
      # _logger.info("Miclassify goal still has attacks to run")
    return finished

  def print_progress(self, criteria, run_counts):
    print("Working on a " + self.__class__.__name__ + " goal.")
    num_below = criteria['correctness'].sum()
    print(str(num_below) + " examples are still correctly classified.")
    if self.new_work_goal is None:
      print("No work goal: running all attacks indefinitely")
    else:
      print("Working until all attacks have been run enough times")
      filtered_run_counts = self.filter(run_counts, criteria)
      filtered_work_before = self.filter(self.work_before, criteria)
      for ac in self.new_work_goal:
        goal = self.new_work_goal[ac]
        new = filtered_run_counts[ac] - filtered_work_before[ac]
        if new.size > 0:
          min_new = new.min()
          if min_new < goal:
            num_min = (new == min_new).sum()
            print("\t" + str(ac) + ": goal of " + str(goal) + " runs, but "
                  + str(num_min) + " examples have been run only " + str(min_new)
                  + " times")

  def filter(self, run_counts, criteria):
    correctness = criteria['correctness']
    assert correctness.dtype == np.bool
    filtered_counts = deep_copy(run_counts)
    for key in filtered_counts:
      filtered_counts[key] = filtered_counts[key][correctness]
    return filtered_counts

  def get_attack_config(self, attack_configs, run_counts, criteria):
    if self.new_work_goal is not None:
      correct_work_before = self.filter(self.work_before, criteria)
      correct_run_counts = self.filter(run_counts, criteria)
      attack_configs = unfinished_attack_configs(self.new_work_goal,
                                                 correct_work_before,
                                                 correct_run_counts)
    attack_config = attack_configs[self.rng.randint(len(attack_configs))]
    return attack_config

  def extra_criteria(self):
    if self.break_ties == "wrong_confidence":
      return ["wrong_confidence"], _WrongConfidenceFactory()
    else:
      raise NotImplementedError()

  def request_examples(self, attack_config, criteria, run_counts, batch_size):
    correctness = criteria['correctness']
    assert correctness.dtype == np.bool
    total = correctness.size
    total_correct = correctness.sum()
    all_idxs = np.arange(total)
    run_counts = run_counts[attack_config]
    if total_correct > 0:
      correct_idxs = all_idxs[correctness]
      assert correct_idxs.size == total_correct
      run_counts = run_counts[correctness]
      pairs = safe_zip(correct_idxs, run_counts)
    else:
      pairs = safe_zip(all_idxs, run_counts)
    # In PY3, pairs is now an iterator.
    # To support sorting, we need to make it a list.
    pairs = list(pairs)

    def key(pair):
      return pair[1]
    pairs.sort(key=key)
    idxs = [pair[0] for pair in pairs]
    while len(idxs) < batch_size:
      needed = batch_size - len(idxs)
      idxs = idxs + idxs[:needed]
    if len(idxs) > batch_size:
      idxs = idxs[:batch_size]
    idxs = np.array(idxs)
    return idxs

  def new_wins(self, orig_criteria, orig_idx, new_criteria, new_idx):
    orig_correct = orig_criteria['correctness'][orig_idx]
    new_correct = new_criteria['correctness'][new_idx]
    if orig_correct and not new_correct:
      return True
    if (not orig_correct) and new_correct:
      return False
    assert orig_correct == new_correct
    if self.break_ties == "wrong_confidence":
      new = new_criteria["wrong_confidence"][new_idx]
      orig = orig_criteria['wrong_confidence'][orig_idx]
      return new > orig
    else:
      raise NotImplementedError(self.break_ties)


class MaxConfidence(AttackGoal):
  """
  The AttackGoal corresponding the MaxConfidence procedure.

  Reference: https://openreview.net/forum?id=H1g0piA9tQ

  This should be used with a recipe that includes AttackConfigs
  that target all of the classes, plus an any additional AttackConfigs
  that may help to avoid gradient masking.

  This AttackGoal prioritizes getting all examples above a specified
  threshold. (If the threshold is set to 1, then no examples are above
  the threshold, so all are attacked equally often). The MaxConfidence
  attack procedure against *a single example* is optimal regardless of
  threshold, so long as t >= 0.5, but when attacking a population of
  examples with finite computation time, knowledge of the threshold is
  necessary to determine which examples to prioritize attacking.

  :param t: Prioritize pushing examples above this threshold.
  :param new_work_goal: Optional dict mapping AttackConfigs to ints.
    The int specifies the number of times to run each AttackConfig on each
    below-threshold example before giving up.
    If not specified, this goal runs all available attacks and never gives
    up.
  """

  def __init__(self, t=1., new_work_goal=None):
    super(MaxConfidence, self).__init__()
    self.t = t
    self.new_work_goal = new_work_goal
    if new_work_goal is not None:
      for key in new_work_goal:
        assert isinstance(key, AttackConfig)
        assert isinstance(new_work_goal[key], int)
    self.rng = np.random.RandomState([2018, 10, 7, 12])

  def filter(self, run_counts, criteria):
    wrong_confidence = criteria['wrong_confidence']
    below_t = wrong_confidence <= self.t
    filtered_counts = deep_copy(run_counts)
    for key in filtered_counts:
      filtered_counts[key] = filtered_counts[key][below_t]
    return filtered_counts

  def extra_criteria(self):
    return ["wrong_confidence"], _WrongConfidenceFactory()

  def is_satisfied(self, criteria, run_counts):
    wrong_confidence = criteria['wrong_confidence']
    if wrong_confidence.min() > self.t:
      _logger.info("Everything is above threshold " + str(self.t))
      _logger.info("Done with MaxConfidence goal")
      return True
    if self.new_work_goal is None:
      return False
    filtered_run_counts = self.filter(run_counts, criteria)
    filtered_work_before = self.filter(self.work_before, criteria)
    unfinished = unfinished_attack_configs(self.new_work_goal,
                                           filtered_work_before,
                                           filtered_run_counts,
                                           log=False)
    finished = len(unfinished) == 0
    if finished:
      _logger.info(
          "MaxConfidence timed out after running all requested attacks")
    else:
      pass
    return finished

  def print_progress(self, criteria, run_counts):
    print("Working on a " + self.__class__.__name__ + " goal.")
    if self.t == 1.:
      print("Threshold of 1, so just driving up confidence of all examples.")
    else:
      print("Target threshold of " + str(self.t))
      num_below = (criteria['wrong_confidence'] <= self.t).sum()
      print(str(num_below) + " examples are below the target threshold.")
    if self.new_work_goal is None:
      print("No work goal: running all attacks indefinitely")
    else:
      print("Working until all attacks have been run enough times")
      filtered_run_counts = self.filter(run_counts, criteria)
      filtered_work_before = self.filter(self.work_before, criteria)
      for ac in self.new_work_goal:
        goal = self.new_work_goal[ac]
        new = filtered_run_counts[ac] - filtered_work_before[ac]
        min_new = new.min()
        if min_new < goal:
          num_min = (new == min_new).sum()
          print("\t" + str(ac) + ": goal of " + str(goal) + " runs, but "
                + str(num_min) + " examples have been run only " + str(min_new)
                + " times")

  def get_attack_config(self, attack_configs, run_counts, criteria):
    # TODO: refactor to avoid this duplicated method
    if self.new_work_goal is not None:
      correct_work_before = self.filter(self.work_before, criteria)
      correct_run_counts = self.filter(run_counts, criteria)
      attack_configs = unfinished_attack_configs(self.new_work_goal,
                                                 correct_work_before,
                                                 correct_run_counts)
    attack_config = attack_configs[self.rng.randint(len(attack_configs))]
    return attack_config

  def start(self, run_counts):
    _logger.info("Started working on a MaxConfidence goal")
    _logger.info("Threshold: " + str(self.t))
    if self.new_work_goal is None:
      if self.t >= 1.:
        _logger.info("This goal will run forever")
      else:
        _logger.info("This goal will run until all examples have confidence"
                     + " greater than " + str(self.t) + ", which may never"
                     + " happen.")
    self.work_before = deep_copy(run_counts)

  def request_examples(self, attack_config, criteria, run_counts, batch_size):
    wrong_confidence = criteria['wrong_confidence']
    below_t = wrong_confidence <= self.t
    assert below_t.dtype == np.bool
    total = below_t.size
    total_below = below_t.sum()
    all_idxs = np.arange(total)
    run_counts = run_counts[attack_config]
    if total_below > 0:
      correct_idxs = all_idxs[below_t]
      assert correct_idxs.size == total_below
      run_counts = run_counts[below_t]
      pairs = safe_zip(correct_idxs, run_counts)
    else:
      pairs = safe_zip(all_idxs, run_counts)

    def key(pair):
      return pair[1]
    pairs.sort(key=key)
    idxs = [pair[0] for pair in pairs]
    while len(idxs) < batch_size:
      needed = batch_size - len(idxs)
      idxs = idxs + idxs[:needed]
    if len(idxs) > batch_size:
      idxs = idxs[:batch_size]
    idxs = np.array(idxs)
    return idxs

  def new_wins(self, orig_criteria, orig_idx, new_criteria, new_idx):
    new_wrong_confidence = new_criteria['wrong_confidence'][new_idx]
    orig_wrong_confidence = orig_criteria['wrong_confidence'][orig_idx]
    return new_wrong_confidence > orig_wrong_confidence


def unfinished_attack_configs(new_work_goal, work_before, run_counts,
                              log=False):
  """
  Returns a list of attack configs that have not yet been run the desired
  number of times.
  :param new_work_goal: dict mapping attacks to desired number of times to run
  :param work_before: dict mapping attacks to number of times they were run
    before starting this new goal. Should be prefiltered to include only
    examples that don't already meet the primary goal
  :param run_counts: dict mapping attacks to total number of times they have
    ever been run. Should be prefiltered to include only examples that don't
    already meet the primary goal
  """

  assert isinstance(work_before, dict), work_before

  for key in work_before:
    value = work_before[key]
    assert value.ndim == 1, value.shape
    if key in run_counts:
      assert run_counts[key].shape == value.shape

  attack_configs = []
  for attack_config in new_work_goal:
    done_now = run_counts[attack_config]
    if log:
      _logger.info(str(attack_config) +
                   " ave run count: " + str(done_now.mean()))
      _logger.info(str(attack_config) +
                   " min run count: " + str(done_now.min()))
    done_before = work_before[attack_config]
    if log:
      _logger.info(str(attack_config) + " mean work before: " +
                   str(done_before.mean()))
    # This is the vector for all examples
    new = done_now - done_before
    # The work is only done when it has been done for every example
    new = new.min()
    assert isinstance(new, (int, np.int64)), type(new)
    new_goal = new_work_goal[attack_config]
    assert isinstance(new_goal, int), type(new_goal)
    if new < new_goal:
      if log:
        _logger.info(str(attack_config) + " has run " +
                     str(new) + " of " + str(new_goal))
      attack_configs.append(attack_config)
  return attack_configs


class _CriteriaFactory(object):
  """
  A factory that builds the expression to evaluate all criteria.
  """

  def __init__(self, model, extra_criteria_factory=None):
    self.model = model
    self.extra_criteria_factory = extra_criteria_factory
    properties_to_hash = (model, )
    if extra_criteria_factory is not None:
      if extra_criteria_factory.properties_to_hash is not None:
        extra_properties = extra_criteria_factory.properties_to_hash
        properties_to_hash = properties_to_hash + extra_properties
    self.properties_to_hash = properties_to_hash

  def __hash__(self):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    return self.properties_to_hash.__hash__()

  def __eq__(self, other):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    if not isinstance(other, _CriteriaFactory):
      return False
    if (type(self.extra_criteria_factory) is not
        type(other.extra_criteria_factory)):
      return False
    return self.properties_to_hash == other.properties_to_hash

  def __call__(self):
    x_batch = self.model.make_input_placeholder()
    y_batch = self.model.make_label_placeholder()

    predictions = self.model.get_probs(x_batch)
    correct = tf.equal(tf.argmax(y_batch, axis=-1),
                       tf.argmax(predictions, axis=-1))
    max_probs = tf.reduce_max(predictions, axis=1)

    if self.extra_criteria_factory is not None:
      extra_criteria = self.extra_criteria_factory(x_batch, y_batch,
                                                   predictions, correct,
                                                   max_probs)
    else:
      extra_criteria = tuple([])

    return (x_batch, y_batch), (correct, max_probs) + extra_criteria


class _ExtraCriteriaFactory(object):
  """
  A factory that builds extra criteria
  """

  def __init__(self, properties_to_hash=None):
    self.properties_to_hash = properties_to_hash

  def __hash__(self):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    return self.properties_to_hash.__hash__()

  def __eq__(self, other):
    # Make factory hashable so that no two factories for the
    # same model will be used to build redundant tf graphs
    if not isinstance(other, _ExtraCriteriaFactory):
      return False
    return self.properties_to_hash == other.properties_to_hash

  def __call__(self, x_batch, y_batch, predictions, correct, max_probs):
    raise NotImplementedError()


class _WrongConfidenceFactory(_ExtraCriteriaFactory):
  def __call__(self, x_batch, y_batch, predictions, correct, max_probs):
    max_wrong_probs = tf.reduce_max(predictions * (1. - y_batch), axis=1)
    return tuple([max_wrong_probs])


def bundle_examples_with_goal(sess, model, adv_x_list, y, goal,
                              report_path):
  """
  A post-processor version of attack bundling, that chooses the strongest
  example from the output of multiple earlier bundling strategies.

  :param sess: tf.session.Session
  :param model: cleverhans.model.Model
  :param adv_x_list: list of numpy arrays
    Each entry in the list is the output of a previous bundler; it is an
      adversarial version of the whole dataset.
  :param y: numpy array containing true labels
  :param goal: AttackGoal to use to choose the best version of each adversarial
    example
  :param report_path: str, the path the report will be saved to
  """

  # Check the input
  num_attacks = len(adv_x_list)
  assert num_attacks > 0
  adv_x_0 = adv_x_list[0]
  assert isinstance(adv_x_0, np.ndarray)
  assert all(adv_x.shape == adv_x_0.shape for adv_x in adv_x_list)

  # Allocate the output
  out = np.zeros_like(adv_x_0)
  m = adv_x_0.shape[0]
  # Initialize with negative sentinel values to make sure everything is
  # written to
  correctness = -np.ones(m, dtype='int32')
  confidence = -np.ones(m, dtype='float32')

  # Gather criteria
  criteria = [goal.get_criteria(sess, model, adv_x, y) for adv_x in adv_x_list]
  assert all('correctness' in c for c in criteria)
  assert all('confidence' in c for c in criteria)
  _logger.info("Accuracy on each advx dataset: ")
  for c in criteria:
    _logger.info("\t" + str(c['correctness'].mean()))

  for example_idx in range(m):
    # Index of the best attack for this example
    attack_idx = 0
    # Find the winner
    for candidate_idx in range(1, num_attacks):
      if goal.new_wins(criteria[attack_idx], example_idx,
                       criteria[candidate_idx], example_idx):
        attack_idx = candidate_idx
    # Copy the winner into the output
    out[example_idx] = adv_x_list[attack_idx][example_idx]
    correctness[example_idx] = criteria[attack_idx]['correctness'][example_idx]
    confidence[example_idx] = criteria[attack_idx]['confidence'][example_idx]

  assert correctness.min() >= 0
  assert correctness.max() <= 1
  assert confidence.min() >= 0.
  assert confidence.max() <= 1.

  report = ConfidenceReport()
  report['bundled'] = ConfidenceReportEntry(correctness, confidence)
  serial.save(report_path, report)
  assert report_path.endswith('.joblib')
  adv_x_path = report_path[:-len('.joblib')] + "_adv_x.npy"
  np.save(adv_x_path, out)

def spsa_max_confidence_recipe(sess, model, x, y, nb_classes, eps,
                               clip_min, clip_max, nb_iter,
                               report_path,
                               spsa_samples=SPSA.DEFAULT_SPSA_SAMPLES,
                               spsa_iters=SPSA.DEFAULT_SPSA_ITERS):
  """Runs the MaxConfidence attack using SPSA as the underlying optimizer.

  Even though this runs only one attack, it must be implemented as a bundler
  because SPSA supports only batch_size=1. The cleverhans.attacks.MaxConfidence
  attack internally multiplies the batch size by nb_classes, so it can't take
  SPSA as a base attacker. Insteader, we must bundle batch_size=1 calls using
  cleverhans.attack_bundling.MaxConfidence.

  References:
  https://openreview.net/forum?id=H1g0piA9tQ

  :param sess: tf.Session
  :param model: cleverhans.model.Model
  :param x: numpy array containing clean example inputs to attack
  :param y: numpy array containing true labels
  :param nb_classes: int, number of classes
  :param eps: float, maximum size of perturbation (measured by max norm)
  :param nb_iter: int, number of iterations for one version of PGD attacks
    (will also run another version with 25X more iterations)
  :param report_path: str, the path that the report will be saved to.
  """
  spsa = SPSA(model, sess)
  spsa_params = {"eps": eps, "clip_min" : clip_min, "clip_max" : clip_max,
                 "nb_iter": nb_iter, "spsa_samples": spsa_samples,
                 "spsa_iters": spsa_iters}
  attack_configs = []
  dev_batch_size = 1 # The only batch size supported by SPSA
  batch_size = num_devices
  ones = tf.ones(dev_batch_size, tf.int32)
  for cls in range(nb_classes):
    cls_params = copy.copy(spsa_params)
    cls_params['y_target'] = tf.to_float(tf.one_hot(ones * cls, nb_classes))
    cls_attack_config = AttackConfig(spsa, cls_params, "spsa_" + str(cls))
    attack_configs.append(cls_attack_config)
  new_work_goal = {config: 1 for config in attack_configs}
  goals = [MaxConfidence(t=1., new_work_goal=new_work_goal)]
  bundle_attacks(sess, model, x, y, attack_configs, goals, report_path,
                 attack_batch_size=batch_size)



"""
This script evaluates trained models that have been saved to the filesystem.
See mnist_tutorial_picklable.py for instructions.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging

import tensorflow as tf
from tensorflow.python.platform import flags

from cleverhans.dataset import MNIST
from cleverhans.attacks import FastGradientMethod
from cleverhans.utils import set_log_level
from cleverhans.utils_tf import model_eval, silence
from cleverhans.serial import load
silence()

FLAGS = flags.FLAGS


def evaluate_model(filepath,
                   train_start=0, train_end=60000, test_start=0,
                   test_end=10000, batch_size=128,
                   testing=False, num_threads=None):
  """
  Run evaluation on a saved model
  :param filepath: path to model to evaluate
  :param train_start: index of first training set example
  :param train_end: index of last training set example
  :param test_start: index of first test set example
  :param test_end: index of last test set example
  :param batch_size: size of evaluation batches
  """

  # Set TF random seed to improve reproducibility
  tf.set_random_seed(1234)

  # Set logging level to see debug information
  set_log_level(logging.INFO)

  # Create TF session
  if num_threads:
    config_args = dict(intra_op_parallelism_threads=1)
  else:
    config_args = {}
  sess = tf.Session(config=tf.ConfigProto(**config_args))

  # Get MNIST test data
  mnist = MNIST(train_start=train_start, train_end=train_end,
                test_start=test_start, test_end=test_end)
  x_train, y_train = mnist.get_set('train')
  x_test, y_test = mnist.get_set('test')

  # Use Image Parameters
  img_rows, img_cols, nchannels = x_train.shape[1:4]
  nb_classes = y_train.shape[1]

  # Define input TF placeholder
  x = tf.placeholder(tf.float32, shape=(None, img_rows, img_cols,
                                        nchannels))
  y = tf.placeholder(tf.float32, shape=(None, nb_classes))

  eval_params = {'batch_size': batch_size}
  fgsm_params = {
      'eps': 0.3,
      'clip_min': 0.,
      'clip_max': 1.
  }

  def do_eval(preds, x_set, y_set, report_key, is_adv=None):
    acc = model_eval(sess, x, y, preds, x_set, y_set, args=eval_params)
    if is_adv is None:
      report_text = None
    elif is_adv:
      report_text = 'adversarial'
    else:
      report_text = 'legitimate'
    if report_text:
      print('Test accuracy on %s examples: %0.4f' % (report_text, acc))

  with sess.as_default():
    model = load(filepath)
  assert len(model.get_params()) > 0

  # Initialize the Fast Gradient Sign Method (FGSM) attack object and
  # graph
  fgsm = FastGradientMethod(model, sess=sess)
  adv_x = fgsm.generate(x, **fgsm_params)
  preds_adv = model.get_logits(adv_x)
  preds = model.get_logits(x)

  # Evaluate the accuracy of the MNIST model on adversarial examples
  do_eval(preds, x_test, y_test, 'train_clean_train_clean_eval', False)
  do_eval(preds_adv, x_test, y_test, 'clean_train_adv_eval', True)


def main(argv=None):
  _, filepath = argv
  evaluate_model(filepath=filepath)


if __name__ == '__main__':
  tf.app.run()



"""
This tutorial shows how to generate adversarial examples using FGSM
and train a model using adversarial training with TensorFlow.
It is very similar to mnist_tutorial_keras_tf.py, which does the same
thing but with a dependence on keras.
The original paper can be found at:
https://arxiv.org/abs/1412.6572
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import numpy as np
import tensorflow as tf
from tensorflow.python.platform import flags

from cleverhans.loss import CrossEntropy
from cleverhans.dataset import MNIST
from cleverhans.utils_tf import model_eval
from cleverhans.train import train
from cleverhans.attacks import FastGradientMethod
from cleverhans.utils import AccuracyReport, set_log_level
from cleverhans_tutorials.tutorial_models import ModelBasicCNN

FLAGS = flags.FLAGS

NB_EPOCHS = 6
BATCH_SIZE = 128
LEARNING_RATE = 0.001
CLEAN_TRAIN = True
BACKPROP_THROUGH_ATTACK = False
NB_FILTERS = 64


def mnist_tutorial(train_start=0, train_end=60000, test_start=0,
                   test_end=10000, nb_epochs=NB_EPOCHS, batch_size=BATCH_SIZE,
                   learning_rate=LEARNING_RATE,
                   clean_train=CLEAN_TRAIN,
                   testing=False,
                   backprop_through_attack=BACKPROP_THROUGH_ATTACK,
                   nb_filters=NB_FILTERS, num_threads=None,
                   label_smoothing=0.1):
  """
  MNIST cleverhans tutorial
  :param train_start: index of first training set example
  :param train_end: index of last training set example
  :param test_start: index of first test set example
  :param test_end: index of last test set example
  :param nb_epochs: number of epochs to train model
  :param batch_size: size of training batches
  :param learning_rate: learning rate for training
  :param clean_train: perform normal training on clean examples only
                      before performing adversarial training.
  :param testing: if true, complete an AccuracyReport for unit tests
                  to verify that performance is adequate
  :param backprop_through_attack: If True, backprop through adversarial
                                  example construction process during
                                  adversarial training.
  :param label_smoothing: float, amount of label smoothing for cross entropy
  :return: an AccuracyReport object
  """

  # Object used to keep track of (and return) key accuracies
  report = AccuracyReport()

  # Set TF random seed to improve reproducibility
  tf.set_random_seed(1234)

  # Set logging level to see debug information
  set_log_level(logging.DEBUG)

  # Create TF session
  if num_threads:
    config_args = dict(intra_op_parallelism_threads=1)
  else:
    config_args = {}
  sess = tf.Session(config=tf.ConfigProto(**config_args))

  # Get MNIST data
  mnist = MNIST(train_start=train_start, train_end=train_end,
                test_start=test_start, test_end=test_end)
  x_train, y_train = mnist.get_set('train')
  x_test, y_test = mnist.get_set('test')

  # Use Image Parameters
  img_rows, img_cols, nchannels = x_train.shape[1:4]
  nb_classes = y_train.shape[1]

  # Define input TF placeholder
  x = tf.placeholder(tf.float32, shape=(None, img_rows, img_cols,
                                        nchannels))
  y = tf.placeholder(tf.float32, shape=(None, nb_classes))

  # Train an MNIST model
  train_params = {
      'nb_epochs': nb_epochs,
      'batch_size': batch_size,
      'learning_rate': learning_rate
  }
  eval_params = {'batch_size': batch_size}
  fgsm_params = {
      'eps': 0.3,
      'clip_min': 0.,
      'clip_max': 1.
  }
  rng = np.random.RandomState([2017, 8, 30])

  def do_eval(preds, x_set, y_set, report_key, is_adv=None):
    acc = model_eval(sess, x, y, preds, x_set, y_set, args=eval_params)
    setattr(report, report_key, acc)
    if is_adv is None:
      report_text = None
    elif is_adv:
      report_text = 'adversarial'
    else:
      report_text = 'legitimate'
    if report_text:
      print('Test accuracy on %s examples: %0.4f' % (report_text, acc))

  if clean_train:
    model = ModelBasicCNN('model1', nb_classes, nb_filters)
    preds = model.get_logits(x)
    loss = CrossEntropy(model, smoothing=label_smoothing)

    def evaluate():
      do_eval(preds, x_test, y_test, 'clean_train_clean_eval', False)

    train(sess, loss, x_train, y_train, evaluate=evaluate,
          args=train_params, rng=rng, var_list=model.get_params())

    # Calculate training error
    if testing:
      do_eval(preds, x_train, y_train, 'train_clean_train_clean_eval')

    # Initialize the Fast Gradient Sign Method (FGSM) attack object and
    # graph
    fgsm = FastGradientMethod(model, sess=sess)
    adv_x = fgsm.generate(x, **fgsm_params)
    preds_adv = model.get_logits(adv_x)

    # Evaluate the accuracy of the MNIST model on adversarial examples
    do_eval(preds_adv, x_test, y_test, 'clean_train_adv_eval', True)

    # Calculate training error
    if testing:
      do_eval(preds_adv, x_train, y_train, 'train_clean_train_adv_eval')

    print('Repeating the process, using adversarial training')

  # Create a new model and train it to be robust to FastGradientMethod
  model2 = ModelBasicCNN('model2', nb_classes, nb_filters)
  fgsm2 = FastGradientMethod(model2, sess=sess)

  def attack(x):
    return fgsm2.generate(x, **fgsm_params)

  loss2 = CrossEntropy(model2, smoothing=label_smoothing, attack=attack)
  preds2 = model2.get_logits(x)
  adv_x2 = attack(x)

  if not backprop_through_attack:
    # For the fgsm attack used in this tutorial, the attack has zero
    # gradient so enabling this flag does not change the gradient.
    # For some other attacks, enabling this flag increases the cost of
    # training, but gives the defender the ability to anticipate how
    # the atacker will change their strategy in response to updates to
    # the defender's parameters.
    adv_x2 = tf.stop_gradient(adv_x2)
  preds2_adv = model2.get_logits(adv_x2)

  def evaluate2():
    # Accuracy of adversarially trained model on legitimate test inputs
    do_eval(preds2, x_test, y_test, 'adv_train_clean_eval', False)
    # Accuracy of the adversarially trained model on adversarial examples
    do_eval(preds2_adv, x_test, y_test, 'adv_train_adv_eval', True)

  # Perform and evaluate adversarial training
  train(sess, loss2, x_train, y_train, evaluate=evaluate2,
        args=train_params, rng=rng, var_list=model2.get_params())

  # Calculate training errors
  if testing:
    do_eval(preds2, x_train, y_train, 'train_adv_train_clean_eval')
    do_eval(preds2_adv, x_train, y_train, 'train_adv_train_adv_eval')

  return report


def main(argv=None):
  from cleverhans_tutorials import check_installation
  check_installation(__file__)

  mnist_tutorial(nb_epochs=FLAGS.nb_epochs, batch_size=FLAGS.batch_size,
                 learning_rate=FLAGS.learning_rate,
                 clean_train=FLAGS.clean_train,
                 backprop_through_attack=FLAGS.backprop_through_attack,
                 nb_filters=FLAGS.nb_filters)


if __name__ == '__main__':
  flags.DEFINE_integer('nb_filters', NB_FILTERS,
                       'Model size multiplier')
  flags.DEFINE_integer('nb_epochs', NB_EPOCHS,
                       'Number of epochs to train model')
  flags.DEFINE_integer('batch_size', BATCH_SIZE,
                       'Size of training batches')
  flags.DEFINE_float('learning_rate', LEARNING_RATE,
                     'Learning rate for training')
  flags.DEFINE_bool('clean_train', CLEAN_TRAIN, 'Train on clean examples')
  flags.DEFINE_bool('backprop_through_attack', BACKPROP_THROUGH_ATTACK,
                    ('If True, backprop through adversarial example '
                     'construction process during adversarial training'))

  tf.app.run()



"""
Unit tests for format checking
"""

from __future__ import print_function


import os
import subprocess

import cleverhans
from cleverhans.devtools.list_files import list_files
from cleverhans.utils import shell_call

# Enter a manual list of files that are allowed to violate PEP8 here
whitelist_pep8 = [
    # This file is broken but could be fixed
    "../examples/multigpu_advtrain/test_attack_multigpu.py"
]

all_py_files = list_files('.py')

def update_whitelist():
  global whitelist_pep8
  # We don't want to test RL-attack because it has so many dependencies
  # not used elsewhere, and pylint wants to import them all
  whitelist_pep8.extend([os.path.relpath(path, cleverhans.__path__[0])
                         for path in all_py_files if "RL-attack" in path])
  # Similarly, we don't want to require robust_vision_benchmark installed
  whitelist_pep8.extend([os.path.relpath(path, cleverhans.__path__[0])
                         for path in all_py_files
                         if "robust_vision_benchmark" in path])
  # Similarly, we don't want to require that cloud be installed
  whitelist_pep8.extend([os.path.relpath(path, cleverhans.__path__[0])
                         for path in all_py_files
                         if "cloud_client" in path])
  # This example has more dependencies too
  whitelist_pep8.extend([os.path.relpath(path, cleverhans.__path__[0])
                         for path in all_py_files
                         if "facenet_adversarial_faces" in path])
  # This too
  whitelist_pep8.extend([os.path.relpath(path, cleverhans.__path__[0])
                         for path in all_py_files
                         if "madry_lab_challenges" in path])


update_whitelist()


whitelist_docstrings = [
]


def test_format_pep8():
  """
  Test if pep8 is respected.
  """
  files_to_check = []
  module_dir = cleverhans.__path__[0]
  for path in all_py_files:
    rel_path = os.path.relpath(path, module_dir)
    if rel_path in whitelist_pep8:
      continue
    else:
      files_to_check.append(path)
  repo_dir = os.path.join(module_dir, os.pardir)
  rcpath = os.path.join(repo_dir, '.pylintrc')
  assert os.path.exists(rcpath)

  # We must run pylint via the command line and subprocess because of
  # problems with the pylint module.
  # The documentation claims you can run it as a python module, but
  # the documentation is wrong: https://github.com/PyCQA/pylint/issues/1870
  # If you run the version described in the linked issue, pylint
  # calls sys.exit once it is done, so it kills the test.

  # Running all files in one pylint command is important for 2 reasons:
  # 1) Correctness: pylint can detect issues that require access to multiple
  #    files, such as cyclic imports
  # 2) Speed: pylint imports modules for deep analysis, so if you run
  #    multiple subprocesses each needs to re-import tensorflow.
  # On Ian's laptop, pylint takes about 10s per file to run on the repo,
  # and there are about 90 files as of the writing of this comment.
  # Running pylint on all files simultaneously takes about 70s, so it
  # is a little better than a 10X speedup.

  # Running multiple jobs in parallel helps but far less than linearly.
  # On Ian's 4-core laptop, running 4 jobs drops the runtime from 70s
  # to 45s.
  # Some of the work is I/O, so it actually makes some sense to run
  # more jobs than cores. On Ian's 4-core laptop, running 8 jobs drops
  # the runtime to 40s.
  # There's a further complication though: I think each job needs to
  # redo imports, so the total amount of work to do increases with
  # the number of jobs. On Ian's laptop, using 64 jobs causes the
  # runtime to increase to 220s. There is not an obvious simple
  # formula like "use one job per CPU core" or "use way more jobs
  # than cores to saturate I/O". For now I'm hoping that 8 will be
  # a reasonable default: it gets good performance on my laptop,
  # and on machines with fewer than 4 cores there should still be
  # a benefit to not being blocked on I/O.

  try:
    shell_call(['pylint', '--rcfile', rcpath, '--jobs', '8'] + files_to_check)
  except subprocess.CalledProcessError as e:
    raise ValueError(e.output.decode("utf-8"))


if __name__ == "__main__":
  test_format_pep8()



"""Loss functions for training models."""
import copy
import json
import os
import warnings

import numpy as np
import tensorflow as tf

from cleverhans.attacks import Attack
from cleverhans.compat import softmax_cross_entropy_with_logits
from cleverhans.model import Model
from cleverhans.utils import safe_zip

try:
  import tensorflow_probability as tfp
  tf_distributions = tfp.distributions
except ImportError:
  tf_distributions = tf.distributions


class Loss(object):
  """
  An abstract interface for loss wrappers that allows flexible control of
  real examples, adversarial examples and labels. These losses are used
  for defenses (during model training).
  """

  def __init__(self, model, hparams=None, attack=None):
    """
    :param model: Model instance, the model on which to apply the loss.
    :param hparams: dict, hyper-parameters for the loss.
    :param attack: cleverhans.attacks.Attack instance
    """
    assert isinstance(model, Model)
    standard = attack is None or isinstance(attack, Attack)
    deprecated = callable(attack)
    if not standard and not deprecated:
      raise TypeError("`attack` must be `None` or `Attack` subclass instance")
    if deprecated:
      warnings.warn("callable attacks are deprecated, switch to an Attack "
                    "subclass. callable attacks will not be supported after "
                    "2019-05-05.")
      class Wrapper(Attack):
        """
        Temporary wrapper class to be removed when deprecated callable
        arguments are removed.

        :param f: a callable object implementing the attack
        """
        def __init__(self, f):
          dummy_model = Model()
          super(Wrapper, self).__init__(model=dummy_model)
          self.f = f

        def generate(self, x):
          return self.f(x)

      attack = Wrapper(attack)
    self.model = model
    self.hparams = hparams
    self.attack = attack

  def save(self, path):
    """Save loss in json format
    """
    json.dump(dict(loss=self.__class__.__name__,
                   params=self.hparams),
              open(os.path.join(path, 'loss.json'), 'wb'))

  def fprop(self, x, y):
    """Forward propagate the loss.
    Loss should be a scalar value, independent of batch size (i.e. use
    reduce_mean over batch axis, don't use reduce_sum or return a tensor).
    Scalar losses are easier to add together, e.g. through `WeightedSum`.
    Mean losses are easier to redistribute across multiple replicas without
    needing to change learning rates, etc.
    :param x: tensor, a batch of inputs.
    :param y: tensor, a batch of outputs (1-hot labels typically).
    """
    raise NotImplementedError


class WeightedSum(Loss):
  """
  A Loss that adds up a weighted sum of other losses.
  """

  def __init__(self, model, terms):
    self.terms = terms

    Loss.__init__(self, model, locals())

  def fprop(self, x, y, **kwargs):
    weights, loss_objects = safe_zip(*self.terms)
    for weight in weights:
      if isinstance(weight, float):
        continue
      if hasattr(weight, 'ndim'):
        assert weight.ndim == 0
        continue
      raise TypeError("weight of %s is not a type that this function "
                      "knows it can accept yet" % str(weight))
    losses = [loss.fprop(x, y, **kwargs) for loss in loss_objects]
    for loss, loss_object in safe_zip(losses, loss_objects):
      if len(loss.get_shape()) > 0:
        raise ValueError("%s.fprop returned a non-scalar value" %
                         str(loss_object))
    terms = [weight * loss for weight, loss in safe_zip(weights, losses)]

    return tf.add_n(terms)


class CrossEntropy(Loss):
  """Cross-entropy loss for a multiclass softmax classifier.
  :param model: Model instance, the model on which to apply the loss.
  :param smoothing: float, amount of label smoothing for cross-entropy.
  :param attack: function, given an input x, return an attacked x'.
  :param pass_y: bool, if True pass y to the attack
  :param adv_coeff: Coefficient to put on the cross-entropy for
    adversarial examples, if adversarial examples are used.
    The coefficient on the cross-entropy for clean examples is
    1. - adv_coeff.
  :param attack_params: dict, keyword arguments passed to `attack.generate`
  """
  def __init__(self, model, smoothing=0., attack=None, pass_y=False,
               adv_coeff=0.5, attack_params=None,
               **kwargs):
    if smoothing < 0 or smoothing > 1:
      raise ValueError('Smoothing must be in [0, 1]', smoothing)
    self.kwargs = kwargs
    Loss.__init__(self, model, locals(), attack)
    self.smoothing = smoothing
    self.adv_coeff = adv_coeff
    self.pass_y = pass_y
    self.attack_params = attack_params

  def fprop(self, x, y, **kwargs):
    kwargs.update(self.kwargs)
    if self.attack is not None:
      attack_params = copy.copy(self.attack_params)
      if attack_params is None:
        attack_params = {}
      if self.pass_y:
        attack_params['y'] = y
      x = x, self.attack.generate(x, **attack_params)
      coeffs = [1. - self.adv_coeff, self.adv_coeff]
      if self.adv_coeff == 1.:
        x = (x[1],)
        coeffs = (coeffs[1],)
    else:
      x = tuple([x])
      coeffs = [1.]
    assert np.allclose(sum(coeffs), 1.)

    # Catching RuntimeError: Variable -= value not supported by tf.eager.
    try:
      y -= self.smoothing * (y - 1. / tf.cast(y.shape[-1], y.dtype))
    except RuntimeError:
      y.assign_sub(self.smoothing * (y - 1. / tf.cast(y.shape[-1],
                                                      y.dtype)))

    logits = [self.model.get_logits(x, **kwargs) for x in x]
    loss = sum(
        coeff * tf.reduce_mean(softmax_cross_entropy_with_logits(labels=y,
                                                                 logits=logit))
        for coeff, logit in safe_zip(coeffs, logits))
    return loss


class MixUp(Loss):
  """Mixup ( https://arxiv.org/abs/1710.09412 )
  :param model: Model instance, the model on which to apply the loss.
  :param beta: float, beta distribution parameter for MixUp.
  """
  def __init__(self, model, beta, **kwargs):
    del kwargs
    Loss.__init__(self, model, locals())
    self.beta = beta

  def fprop(self, x, y, **kwargs):
    with tf.device('/CPU:0'):
      # Prevent error complaining GPU kernels unavailable for this.
      mix = tf_distributions.Beta(self.beta, self.beta)
      mix = mix.sample([tf.shape(x)[0]] + [1] * (len(x.shape) - 1))
    mix = tf.maximum(mix, 1 - mix)
    mix_label = tf.reshape(mix, [-1, 1])
    xm = x + mix * (x[::-1] - x)
    ym = y + mix_label * (y[::-1] - y)
    logits = self.model.get_logits(xm, **kwargs)
    loss = tf.reduce_mean(softmax_cross_entropy_with_logits(labels=ym,
                                                            logits=logits))
    return loss


class FeaturePairing(Loss):
  """Feature pairing loss.
  :param model: Model instance, the model on which to apply the loss.
  :param weight: float, with of logic pairing loss.
  :param attack: function, given an input x, return an attacked x'.
  """

  def __init__(self, model, weight, attack, **kwargs):
    del kwargs
    Loss.__init__(self, model, locals(), attack)
    self.weight = weight

  def fprop(self, x, y, **kwargs):
    x_adv = self.attack.generate(x)
    d1 = self.model.fprop(x, **kwargs)
    d2 = self.model.fprop(x_adv, **kwargs)
    pairing_loss = [tf.reduce_mean(tf.square(a - b))
                    for a, b in
                    zip(d1[Model.O_FEATURES], d2[Model.O_FEATURES])]
    pairing_loss = tf.reduce_mean(pairing_loss)
    loss = tf.reduce_mean(softmax_cross_entropy_with_logits(
        labels=y, logits=d1[Model.O_LOGITS]))
    loss += tf.reduce_mean(softmax_cross_entropy_with_logits(
        labels=y, logits=d2[Model.O_LOGITS]))
    return loss + self.weight * pairing_loss


class WeightDecay(Loss):
  """Weight decay"""
  def fprop(self, x, y, **kwargs):
    terms = [tf.nn.l2_loss(param)
             for param in self.model.get_params()
             if len(param.get_shape()) > 1]
    out = tf.add_n(terms)
    assert len(out.get_shape()) == 0
    return out


class LossCrossEntropy(Loss):
  """
  Deprecated version of `CrossEntropy` that returns per-example loss rather
  than mean loss.
  """

  def __init__(self, model, smoothing=0., attack=None, **kwargs):
    """Constructor.
    :param model: Model instance, the model on which to apply the loss.
    :param smoothing: float, amount of label smoothing for cross-entropy.
    :param attack: function, given an input x, return an attacked x'.
    """
    if smoothing < 0 or smoothing > 1:
      raise ValueError('Smoothing must be in [0, 1]', smoothing)
    del kwargs
    Loss.__init__(self, model, locals(), attack)
    self.smoothing = smoothing

  def fprop(self, x, y, **kwargs):
    if self.attack is not None:
      x = x, self.attack(x)
    else:
      x = tuple([x])

    # Catching RuntimeError: Variable -= value not supported by tf.eager.
    try:
      y -= self.smoothing * (y - 1. / tf.cast(y.shape[-1], tf.float32))
    except RuntimeError:
      y.assign_sub(self.smoothing * (y - 1. / tf.cast(y.shape[-1],
                                                      tf.float32)))

    logits = [self.model.get_logits(x, **kwargs) for x in x]
    loss = sum(
        softmax_cross_entropy_with_logits(labels=y,
                                          logits=logit)
        for logit in logits)
    warnings.warn("LossCrossEntropy is deprecated, switch to "
                  "CrossEntropy. LossCrossEntropy may be removed on "
                  "or after 2019-03-06.")
    return loss


class LossFeaturePairing(Loss):
  """Deprecated version of `FeaturePairing` that returns per-example loss
  rather than mean loss."""

  def __init__(self, model, weight, attack, **kwargs):
    """Constructor.
    :param model: Model instance, the model on which to apply the loss.
    :param weight: float, with of logic pairing loss.
    :param attack: function, given an input x, return an attacked x'.
    """
    del kwargs
    Loss.__init__(self, model, locals(), attack)
    self.weight = weight

  def fprop(self, x, y, **kwargs):
    x_adv = self.attack(x)
    d1 = self.model.fprop(x, **kwargs)
    d2 = self.model.fprop(x_adv, **kwargs)
    pairing_loss = [tf.reduce_mean(tf.square(a - b))
                    for a, b in
                    zip(d1[Model.O_FEATURES], d2[Model.O_FEATURES])]
    pairing_loss = tf.reduce_mean(pairing_loss)
    loss = softmax_cross_entropy_with_logits(
        labels=y, logits=d1[Model.O_LOGITS])
    loss += softmax_cross_entropy_with_logits(
        labels=y, logits=d2[Model.O_LOGITS])
    warnings.warn("LossFeaturePairing is deprecated, switch to "
                  "FeaturePairing. LossFeaturePairing may be removed "
                  "on or after 2019-03-06.")
    return loss + self.weight * pairing_loss


class LossMixUp(Loss):
  """Deprecated version of `MixUp` that returns per-example loss
  rather than mean loss."""

  def __init__(self, model, beta, **kwargs):
    """Constructor.
    :param model: Model instance, the model on which to apply the loss.
    :param beta: float, beta distribution parameter for MixUp.
    """
    del kwargs
    Loss.__init__(self, model, locals())
    self.beta = beta

  def fprop(self, x, y, **kwargs):
    mix = tf_distributions.Beta(self.beta, self.beta)
    mix = mix.sample([tf.shape(x)[0]] + [1] * (len(x.shape) - 1))
    xm = x + mix * (x[::-1] - x)
    ym = y + mix * (y[::-1] - y)
    logits = self.model.get_logits(xm, **kwargs)
    loss = softmax_cross_entropy_with_logits(labels=ym, logits=logits)
    warnings.warn("LossMixUp is deprecated, switch to "
                  "MixUp. LossFeaturePairing may be removed "
                  "on or after 2019-03-06.")
    return loss



"""
This tutorial shows how to generate adversarial examples using FGSM
and train a model using adversarial training with TensorFlow Eager.

It is similar to mnist_tutorial_tf.py.
mnist_tutorial_tf - dynaminc eager execution.
mnist_tutorial_tf - graph based execution.
The original paper can be found at:
https://arxiv.org/abs/1412.6572
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import numpy as np
import tensorflow as tf
from tensorflow.python.platform import flags

from cleverhans.utils import AccuracyReport
from cleverhans.utils_tfe import train, model_eval
from cleverhans.attacks_tfe import BasicIterativeMethod
from cleverhans.attacks_tfe import FastGradientMethod
from cleverhans.dataset import MNIST
from cleverhans.utils import set_log_level
from cleverhans_tutorials.tutorial_models_tfe import ModelBasicCNNTFE

if tf.executing_eagerly() is True:
  print('TF Eager Activated.')
else:
  raise Exception("Error Enabling Eager Execution.")
tfe = tf.contrib.eager

FLAGS = flags.FLAGS

NB_EPOCHS = 6
BATCH_SIZE = 128
LEARNING_RATE = .001
NB_FILTERS = 64

# Keeps track of implemented attacks.
# Maps attack string taken from bash to attack class
# -- 'fgsm' : FastGradientMethod
# -- 'bim': BasicIterativeMethod

AVAILABLE_ATTACKS = {
    'fgsm': FastGradientMethod,
    'bim': BasicIterativeMethod
}


def attack_selection(attack_string):
  """
  Selects the Attack Class using string input.
  :param attack_string: adversarial attack name in string format
  :return: attack class defined in cleverhans.attacks_eager
  """

  # List of Implemented attacks
  attacks_list = AVAILABLE_ATTACKS.keys()

  # Checking for  requested attack in list of available attacks.
  if attack_string is None:
    raise AttributeError("Attack type is not specified, "
                         "list of available attacks\t".join(attacks_list))
  if attack_string not in attacks_list:
    raise AttributeError("Attack not available "
                         "list of available attacks\t".join(attacks_list))
  # Mapping attack from string to class.
  attack_class = AVAILABLE_ATTACKS[attack_string]
  return attack_class


def mnist_tutorial(train_start=0, train_end=60000, test_start=0,
                   test_end=10000, nb_epochs=NB_EPOCHS, batch_size=BATCH_SIZE,
                   learning_rate=LEARNING_RATE,
                   clean_train=True,
                   testing=False,
                   backprop_through_attack=False,
                   nb_filters=NB_FILTERS, num_threads=None,
                   attack_string=None):
  """
  MNIST cleverhans tutorial
  :param train_start: index of first training set example.
  :param train_end: index of last training set example.
  :param test_start: index of first test set example.
  :param test_end: index of last test set example.
  :param nb_epochs: number of epochs to train model.
  :param batch_size: size of training batches.
  :param learning_rate: learning rate for training.
  :param clean_train: perform normal training on clean examples only
                      before performing adversarial training.
  :param testing: if true, complete an AccuracyReport for unit tests
                  to verify that performance is adequate.
  :param backprop_through_attack: If True, backprop through adversarial
                                  example construction process during
                                  adversarial training.
  :param nb_filters: number of filters in the CNN used for training.
  :param num_threads: number of threads used for running the process.
  :param attack_string: attack name for crafting adversarial attacks and
                          adversarial training, in string format.
  :return: an AccuracyReport object
  """

  # Object used to keep track of (and return) key accuracies
  report = AccuracyReport()

  # Set TF random seed to improve reproducibility
  tf.set_random_seed(1234)

  # Set logging level to see debug information
  set_log_level(logging.DEBUG)

  # Get MNIST test data
  mnist = MNIST(train_start=train_start, train_end=train_end,
                test_start=test_start, test_end=test_end)
  X_train, Y_train = mnist.get_set('train')
  X_test, Y_test = mnist.get_set('test')

  # Use label smoothing
  assert Y_train.shape[1] == 10
  label_smooth = .1
  Y_train = Y_train.clip(label_smooth / 9., 1. - label_smooth)

  # Train an MNIST model
  train_params = {
      'nb_epochs': nb_epochs,
      'batch_size': batch_size,
      'learning_rate': learning_rate
  }

  # Initialize the attack object
  attack_class = attack_selection(attack_string)
  attack_params = {'eps': 0.3, 'clip_min': 0.,
                   'clip_max': 1.}

  rng = np.random.RandomState([2018, 6, 18])
  if clean_train:
    model = ModelBasicCNNTFE(nb_filters=nb_filters)

    def evaluate_clean():
      """Evaluate the accuracy of the MNIST model on legitimate test
      examples
      """
      eval_params = {'batch_size': batch_size}
      acc = model_eval(model, X_test, Y_test, args=eval_params)
      report.clean_train_clean_eval = acc
      assert X_test.shape[0] == test_end - test_start, X_test.shape
      print('Test accuracy on legitimate examples: %0.4f' % acc)

    train(model, X_train, Y_train, evaluate=evaluate_clean,
          args=train_params, rng=rng, var_list=model.get_params())

    if testing:
      # Calculate training error
      eval_params = {'batch_size': batch_size}
      acc = model_eval(model, X_train, Y_train, args=eval_params)
      report.train_clean_train_clean_eval = acc

    # Evaluate the accuracy of the MNIST model on adversarial examples
    eval_par = {'batch_size': batch_size}
    attack = attack_class(model)
    acc = model_eval(
        model, X_test, Y_test, args=eval_par,
        attack=attack, attack_args=attack_params)
    print('Test accuracy on adversarial examples: %0.4f\n' % acc)
    report.clean_train_adv_eval = acc

    # Calculate training error
    if testing:
      eval_par = {'batch_size': batch_size}
      acc = model_eval(
          model, X_train, Y_train, args=eval_par,
          attack=attack, attack_args=attack_params)
      print('Train accuracy on adversarial examples: %0.4f\n' % acc)
      report.train_clean_train_adv_eval = acc

    attack = None
    print("Repeating the process, using adversarial training")

  model_adv_train = ModelBasicCNNTFE(nb_filters=nb_filters)
  attack = attack_class(model_adv_train)

  def evaluate_adv():
    # Accuracy of adversarially trained model on legitimate test inputs
    eval_params = {'batch_size': batch_size}
    accuracy = model_eval(
        model_adv_train, X_test, Y_test,
        args=eval_params)
    print('Test accuracy on legitimate examples: %0.4f' % accuracy)
    report.adv_train_clean_eval = accuracy
    # Accuracy of the adversarially trained model on adversarial examples
    accuracy = model_eval(
        model_adv_train, X_test, Y_test,
        args=eval_params, attack=attack,
        attack_args=attack_params)
    print('Test accuracy on adversarial examples: %0.4f' % accuracy)
    report.adv_train_adv_eval = accuracy

  # Perform and evaluate adversarial training
  train(model_adv_train, X_train, Y_train, evaluate=evaluate_adv,
        args=train_params, rng=rng,
        var_list=model_adv_train.get_params(),
        attack=attack, attack_args=attack_params)

  # Calculate training errors
  if testing:
    eval_params = {'batch_size': batch_size}
    accuracy = model_eval(
        model_adv_train, X_train, Y_train, args=eval_params,
        attack=None, attack_args=None)
    report.train_adv_train_clean_eval = accuracy
    accuracy = model_eval(
        model_adv_train, X_train, Y_train, args=eval_params,
        attack=attack, attack_args=attack_params)
    report.train_adv_train_adv_eval = accuracy
  return report


def main(argv=None):
  from cleverhans_tutorials import check_installation
  check_installation(__file__)

  mnist_tutorial(
      nb_epochs=FLAGS.nb_epochs, batch_size=FLAGS.batch_size,
      learning_rate=FLAGS.learning_rate, clean_train=FLAGS.clean_train,
      backprop_through_attack=FLAGS.backprop_through_attack,
      nb_filters=FLAGS.nb_filters, attack_string=FLAGS.attack)


if __name__ == '__main__':
  flags.DEFINE_integer('nb_filters', NB_FILTERS,
                       'Model size multiplier')
  flags.DEFINE_integer('nb_epochs', NB_EPOCHS,
                       'Number of epochs to train model')
  flags.DEFINE_integer('batch_size', BATCH_SIZE, 'Size of training batches')
  flags.DEFINE_float('learning_rate', LEARNING_RATE,
                     'Learning rate for training')
  flags.DEFINE_bool('clean_train', True, 'Train on clean examples')
  flags.DEFINE_bool('backprop_through_attack', False,
                    ('If True, backprop through adversarial example '
                     'construction process during adversarial training'))
  flags.DEFINE_string('attack', 'fgsm',
                      'Adversarial attack crafted and used for training')
  tf.app.run()



"""File to read in the weights of the neural network from a checkpoint.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import numpy as np
import tensorflow as tf


def read_weights(checkpoint, model_json):
  """Function to read the weights from checkpoint based on json description.

  Args:
    checkpoint: tensorflow checkpoint with trained model to verify
    model_json: path of json file with model description of the network list
      of dictionary items for each layer containing 'type', 'weight_var',
      'bias_var' and 'is_transpose' 'type'is one of {'ff', 'ff_relu' or
      'conv'}; 'weight_var' is the name of tf variable for weights of layer i;
      'bias_var' is the name of tf variable for bias of layer i;
      'is_transpose' is set to True if the weights have to be transposed as
      per convention
      Note that last layer is always feedforward (verification operates at the
      layer below the final softmax for more numerical stability)

  Returns:
    net_weights:
    net_biases:
    net_layer_types: all are lists containing numpy
    version of weights

  Raises:
    ValueError: If layer_types are invalid or variable names not found in
    checkpoint
  """
  # Load checkpoint
  reader = tf.train.load_checkpoint(checkpoint)
  variable_map = reader.get_variable_to_shape_map()
  checkpoint_variable_names = variable_map.keys()
  # Parse JSON file for names
  with tf.gfile.Open(model_json) as f:
    list_model_var = json.load(f)

  net_layer_types = []
  net_weights = []
  net_biases = []

  # Checking validity of the input and adding to list
  for layer_model_var in list_model_var:
    if layer_model_var['type'] not in {'ff', 'ff_relu', 'conv'}:
      raise ValueError('Invalid layer type in description')
    if (layer_model_var['weight_var'] not in checkpoint_variable_names or
        layer_model_var['bias_var'] not in checkpoint_variable_names):
      raise ValueError('Variable names not found in checkpoint')
    net_layer_types.append(layer_model_var['type'])
    layer_weight = reader.get_tensor(layer_model_var['weight_var'])
    layer_bias = reader.get_tensor(layer_model_var['bias_var'])
    # TODO(aditirag): is there a way to automatically check when to transpose
    # We want weights W such that x^{i+1} = W^i x^i + b^i
    # Can think of a hack involving matching shapes but if shapes are equal
    # it can be ambiguous
    if layer_model_var['is_transpose']:
      layer_weight = np.transpose(layer_weight)
    net_weights.append(layer_weight)
    net_biases.append(np.reshape(layer_bias, (np.size(layer_bias), 1)))
  return net_weights, net_biases, net_layer_types



"""
Functionality for showing images in pyplot.
See also cleverhans.plot.image for loading/saving image files, showing
images in 3rd party viewers, etc.
"""
import numpy as np
from six.moves import range

def pair_visual(original, adversarial, figure=None):
  """
  This function displays two images: the original and the adversarial sample
  :param original: the original input
  :param adversarial: the input after perturbations have been applied
  :param figure: if we've already displayed images, use the same plot
  :return: the matplot figure to reuse for future samples
  """
  import matplotlib.pyplot as plt

  # Squeeze the image to remove single-dimensional entries from array shape
  original = np.squeeze(original)
  adversarial = np.squeeze(adversarial)

  # Ensure our inputs are of proper shape
  assert(len(original.shape) == 2 or len(original.shape) == 3)

  # To avoid creating figures per input sample, reuse the sample plot
  if figure is None:
    plt.ion()
    figure = plt.figure()
    figure.canvas.set_window_title('Cleverhans: Pair Visualization')

  # Add the images to the plot
  perturbations = adversarial - original
  for index, image in enumerate((original, perturbations, adversarial)):
    figure.add_subplot(1, 3, index + 1)
    plt.axis('off')

    # If the image is 2D, then we have 1 color channel
    if len(image.shape) == 2:
      plt.imshow(image, cmap='gray')
    else:
      plt.imshow(image)

    # Give the plot some time to update
    plt.pause(0.01)

  # Draw the plot and return
  plt.show()
  return figure

def grid_visual(data):
  """
  This function displays a grid of images to show full misclassification
  :param data: grid data of the form;
      [nb_classes : nb_classes : img_rows : img_cols : nb_channels]
  :return: if necessary, the matplot figure to reuse
  """
  import matplotlib.pyplot as plt

  # Ensure interactive mode is disabled and initialize our graph
  plt.ioff()
  figure = plt.figure()
  figure.canvas.set_window_title('Cleverhans: Grid Visualization')

  # Add the images to the plot
  num_cols = data.shape[0]
  num_rows = data.shape[1]
  num_channels = data.shape[4]
  for y in range(num_rows):
    for x in range(num_cols):
      figure.add_subplot(num_rows, num_cols, (x + 1) + (y * num_cols))
      plt.axis('off')

      if num_channels == 1:
        plt.imshow(data[x, y, :, :, 0], cmap='gray')
      else:
        plt.imshow(data[x, y, :, :, :])

  # Draw the plot and return
  plt.show()
  return figure


def get_logits_over_interval(sess, model, x_data, fgsm_params,
                             min_epsilon=-10., max_epsilon=10.,
                             num_points=21):
  """Get logits when the input is perturbed in an interval in adv direction.

  Args:
      sess: Tf session
      model: Model for which we wish to get logits.
      x_data: Numpy array corresponding to single data.
              point of shape [height, width, channels].
      fgsm_params: Parameters for generating adversarial examples.
      min_epsilon: Minimum value of epsilon over the interval.
      max_epsilon: Maximum value of epsilon over the interval.
      num_points: Number of points used to interpolate.

  Returns:
      Numpy array containing logits.

  Raises:
      ValueError if min_epsilon is larger than max_epsilon.
  """
  # Get the height, width and number of channels
  height = x_data.shape[0]
  width = x_data.shape[1]
  channels = x_data.shape[2]

  x_data = np.expand_dims(x_data, axis=0)
  import tensorflow as tf
  from cleverhans.attacks import FastGradientMethod

  # Define the data placeholder
  x = tf.placeholder(dtype=tf.float32,
                     shape=[1, height,
                            width,
                            channels],
                     name='x')
  # Define adv_x
  fgsm = FastGradientMethod(model, sess=sess)
  adv_x = fgsm.generate(x, **fgsm_params)

  if min_epsilon > max_epsilon:
    raise ValueError('Minimum epsilon is less than maximum epsilon')

  eta = tf.nn.l2_normalize(adv_x - x, dim=0)
  epsilon = tf.reshape(tf.lin_space(float(min_epsilon),
                                    float(max_epsilon),
                                    num_points),
                       (num_points, 1, 1, 1))
  lin_batch = x + epsilon * eta
  logits = model.get_logits(lin_batch)
  with sess.as_default():
    log_prob_adv_array = sess.run(logits,
                                  feed_dict={x: x_data})
  return log_prob_adv_array

def linear_extrapolation_plot(log_prob_adv_array, y, file_name,
                              min_epsilon=-10, max_epsilon=10,
                              num_points=21):
  """Generate linear extrapolation plot.

  Args:
      log_prob_adv_array: Numpy array containing log probabilities
      y: Tf placeholder for the labels
      file_name: Plot filename
      min_epsilon: Minimum value of epsilon over the interval
      max_epsilon: Maximum value of epsilon over the interval
      num_points: Number of points used to interpolate
  """
  import matplotlib
  matplotlib.use('Agg')
  import matplotlib.pyplot as plt

  figure = plt.figure()
  figure.canvas.set_window_title('Cleverhans: Linear Extrapolation Plot')

  correct_idx = np.argmax(y, axis=0)
  fig = plt.figure()
  plt.xlabel('Epsilon')
  plt.ylabel('Logits')
  x_axis = np.linspace(min_epsilon, max_epsilon, num_points)
  plt.xlim(min_epsilon - 1, max_epsilon + 1)
  for i in range(y.shape[0]):
    if i == correct_idx:
      ls = '-'
      linewidth = 5
    else:
      ls = '--'
      linewidth = 2
    plt.plot(
        x_axis,
        log_prob_adv_array[:, i],
        ls=ls,
        linewidth=linewidth,
        label='{}'.format(i))
  plt.legend(loc='best', fontsize=14)
  plt.show()
  fig.savefig(file_name)
  plt.clf()
  return figure



"""
Attacks for TensorFlow Eager
"""
from distutils.version import LooseVersion

import numpy as np
import tensorflow as tf

from cleverhans import attacks
from cleverhans import utils
from cleverhans.model import CallableModelWrapper, wrapper_warning
from cleverhans.model import Model
from cleverhans.loss import LossCrossEntropy

_logger = utils.create_logger("cleverhans.attacks_tfe")


if LooseVersion(tf.__version__) < LooseVersion('1.8.0'):
  error_msg = ('For eager execution',
               'use Tensorflow version greather than 1.8.0.')
  raise ValueError(error_msg)


class Attack(attacks.Attack):
  """
  Abstract base class for all eager attack classes.
  :param model: An instance of the cleverhans.model.Model class.
  :param back: The backend to use. Inherited from AttackBase class.
  :param dtypestr: datatype of the input data samples and crafted
                   adversarial attacks.
  """

  def __init__(self, model, dtypestr='float32'):
    super(Attack, self).__init__(model, dtypestr=dtypestr)
    # Validate the input arguments.
    if dtypestr != 'float32' and dtypestr != 'float64':
      raise ValueError("Unexpected input for argument dtypestr.")
    self.tf_dtype = tf.as_dtype(dtypestr)
    self.np_dtype = np.dtype(dtypestr)

    if not isinstance(model, Model):
      raise ValueError("The model argument should be an instance of"
                       " the cleverhans.model.Model class.")
    # Prepare attributes
    self.model = model
    self.dtypestr = dtypestr

  def construct_graph(self, **kwargs):
    """
    Constructs the graph required to run the attacks.
    Is inherited from the attack class, is overloaded
    to raise an error.
    """
    error = "This method is not required for eager execution."
    raise AttributeError(error)

  def generate_np(self, x_val, **kwargs):
    """
    Generate adversarial examples and return them as a NumPy array.

    :param x_val: A NumPy array with the original inputs.
    :param **kwargs: optional parameters used by child classes.
    :return: A NumPy array holding the adversarial examples.
    """
    tfe = tf.contrib.eager
    x = tfe.Variable(x_val)
    adv_x = self.generate(x, **kwargs)
    return adv_x.numpy()

  def construct_variables(self, kwargs):
    """
    Construct the inputs to the attack graph.
    Is inherited from the attack class, is overloaded
    to raise an error.
    """
    error = "This method is not required for eager execution."
    raise AttributeError(error)


class FastGradientMethod(Attack, attacks.FastGradientMethod):
  """
  Inherited class from Attack and cleverhans.attacks.FastGradientMethod.

  This attack was originally implemented by Goodfellow et al. (2015) with the
  infinity norm (and is known as the "Fast Gradient Sign Method"). This
  implementation extends the attack to other norms, and is therefore called
  the Fast Gradient Method.
  Paper link: https://arxiv.org/abs/1412.6572
  """

  def __init__(self, model, dtypestr='float32', **kwargs):
    """
    Creates a FastGradientMethod instance in eager execution.
    :model: cleverhans.model.Model
    :dtypestr: datatype in the string format.
    """
    del kwargs
    if not isinstance(model, Model):
      wrapper_warning()
      model = CallableModelWrapper(model, 'probs')

    super(FastGradientMethod, self).__init__(model, dtypestr)

  def generate(self, x, **kwargs):
    """
    Generates the adversarial sample for the given input.
    :param x: The model's inputs.
    :param eps: (optional float) attack step size (input variation)
    :param ord: (optional) Order of the norm (mimics NumPy).
                Possible values: np.inf, 1 or 2.
    :param y: (optional) A tf variable` with the model labels. Only provide
              this parameter if you'd like to use true labels when crafting
              adversarial samples. Otherwise, model predictions are used as
              labels to avoid the "label leaking" effect (explained in this
              paper: https://arxiv.org/abs/1611.01236). Default is None.
              Labels should be one-hot-encoded.
    :param y_target: (optional) A tf variable` with the labels to target.
                        Leave y_target=None if y is also set.
                        Labels should be one-hot-encoded.
    :param clip_min: (optional float) Minimum input component value
    :param clip_max: (optional float) Maximum input component value
    """
    # Parse and save attack-specific parameters
    assert self.parse_params(**kwargs)
    labels, _nb_classes = self.get_or_guess_labels(x, kwargs)
    return self.fgm(x, labels=labels, targeted=(self.y_target is not None))

  def fgm(self, x, labels, targeted=False):
    """
    TensorFlow Eager implementation of the Fast Gradient Method.
    :param x: the input variable
    :param targeted: Is the attack targeted or untargeted? Untargeted, the
                     default, will try to make the label incorrect.
                     Targeted will instead try to move in the direction
                     of being more like y.
    :return: a tensor for the adversarial example
    """
    # Compute loss
    with tf.GradientTape() as tape:
      # input should be watched because it may be
      # combination of trainable and non-trainable variables
      tape.watch(x)
      loss_obj = LossCrossEntropy(self.model, smoothing=0.)
      loss = loss_obj.fprop(x=x, y=labels)
      if targeted:
        loss = -loss

    # Define gradient of loss wrt input
    grad = tape.gradient(loss, x)
    optimal_perturbation = attacks.optimize_linear(grad, self.eps, self.ord)

    # Add perturbation to original example to obtain adversarial example
    adv_x = x + optimal_perturbation

    # If clipping is needed
    # reset all values outside of [clip_min, clip_max]
    if (self.clip_min is not None) and (self.clip_max is not None):
      adv_x = tf.clip_by_value(adv_x, self.clip_min, self.clip_max)
    return adv_x


class BasicIterativeMethod(Attack, attacks.BasicIterativeMethod):
  """
  Inherited class from Attack and cleverhans.attacks.BasicIterativeMethod.

  The Basic Iterative Method (Kurakin et al. 2016). The original paper used
  hard labels for this attack; no label smoothing.
  Paper link: https://arxiv.org/pdf/1607.02533.pdf
  """

  FGM_CLASS = FastGradientMethod

  def __init__(self, model, dtypestr='float32'):
    """
    Creates a BasicIterativeMethod instance in eager execution.
    :param model: cleverhans.model.Model
    :param dtypestr: datatype in the string format.
    """
    if not isinstance(model, Model):
      wrapper_warning()
      model = CallableModelWrapper(model, 'probs')

    super(BasicIterativeMethod, self).__init__(model, dtypestr)



"""Utility functions for mocking up tests.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy

from keras.utils import np_utils
import numpy as np

from cleverhans.dataset import Dataset


def random_feed_dict(rng, placeholders):
  """
  Returns random data to be used with `feed_dict`.
  :param rng: A numpy.random.RandomState instance
  :param placeholders: List of tensorflow placeholders
  :return: A dict mapping placeholders to random numpy values
  """

  output = {}

  for placeholder in placeholders:
    if placeholder.dtype != 'float32':
      raise NotImplementedError()
    value = rng.randn(*placeholder.shape).astype('float32')
    output[placeholder] = value

  return output

class SimpleDataset(Dataset):
  """
  A dataset containing random values.
  Values are uniformly distributed, either in [0, max_val] or [-1, max_val].
  """

  def __init__(self, dim=2, train_start=0, train_end=3, test_start=0, test_end=5, center=False, max_val=1.,
               nb_classes=5):
    kwargs = copy.copy(locals())
    del kwargs['self']
    if "__class__" in kwargs:
      del kwargs["__class__"]
    super(SimpleDataset, self).__init__(kwargs)
    self.__dict__.update(kwargs)
    train_x_rng = np.random.RandomState([2018, 11, 9, 1])
    # Even if train_start is not 0, we should still generate the first training examples from the rng.
    # This way the dataset looks like it is an array of deterministic data that we index using train_start.
    self.x_train = train_x_rng.uniform(- center * max_val, max_val, (train_end, dim))[train_start:]
    # Use a second rng for the test set so that it also looks like an array of deterministic data that we
    # index into, unaffected by the number of training examples.
    test_x_rng = np.random.RandomState([2018, 11, 9, 2])
    self.x_test = test_x_rng.uniform(- center * max_val, max_val, (test_end, dim))[test_start:]
    # Likewise, to keep the number of examples read from the rng affecting the values of the labels, we
    # must generate the labels from a different rng
    train_y_rng = np.random.RandomState([2018, 11, 9, 3])
    self.y_train = train_y_rng.randint(low=0, high=nb_classes, size=(train_end, 1))[train_start:]
    test_y_rng = np.random.RandomState([2018, 11, 9, 4])
    self.y_test = test_y_rng.randint(low=0, high=nb_classes, size=(test_end, 1))[test_start:]
    assert self.x_train.shape[0] == self.y_train.shape[0]
    assert self.x_test.shape[0] == self.y_test.shape[0]
    self.y_train = np_utils.to_categorical(self.y_train, nb_classes)
    self.y_test = np_utils.to_categorical(self.y_test, nb_classes)



"""Functionality for making confidence reports.

A confidence report is a dictionary.
Each dictionary key is the name of a type of data:
  clean : Clean data
  bundled : bundled adversarial examples
Each value in the dictionary contains an array of bools indicating whether
the model got each example correct and an array containing the confidence
that the model assigned to each prediction.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import OrderedDict
import logging
import time
import warnings

import numpy as np
import six
import tensorflow as tf

from cleverhans.attacks import MaxConfidence
from cleverhans.attacks import Semantic
from cleverhans.evaluation import correctness_and_confidence
from cleverhans.evaluation import run_attack
from cleverhans.utils import set_log_level
from cleverhans.serial import load, save
from cleverhans.utils_tf import infer_devices

# Defaults. Imported elsewhere so that command line script defaults match
# function defaults.
TRAIN_START = 0
TRAIN_END = 60000
TEST_START = 0
TEST_END = 10000
WHICH_SET = 'test'
RECIPE = 'basic_max_confidence_recipe'
REPORT_PATH = None
# Used for `make_confidence_report` but not `make_confidence_report_bundled`
devices = infer_devices()
num_devices = len(devices)
BATCH_SIZE = 128 * num_devices
MC_BATCH_SIZE = 16 * num_devices
NB_ITER = 40
BASE_EPS_ITER = None  # Differs by dataset
SAVE_ADVX = 1


class ConfidenceReport(OrderedDict):
  """
  A data structure reporting how much confidence a model assigned to its
  predictions on each example and whether those predictions were correct.
  This class is just a dictionary with some type checks.
  It maps string data type names (like "clean" for clean data or "Semantic"
  for semantic adversarial examples) to ConfidenceReportEntry instances.

  :param iterable: optional iterable containing (key, value) tuples
  """

  def __init__(self, iterable=None):
    super(ConfidenceReport, self).__init__()
    # This field tracks whether the report is completed.
    # It's important e.g. for reports that are made by bundlers and repeatedly
    # written to disk during the process. This field makes it possible to tell
    # whether a report on disk is complete or whether the bundling process
    # got killed (e.g. due to VM migration)
    self.completed = False
    if iterable is not None:
      # pickle sometimes wants to use this interface to unpickle the OrderedDict
      for key, value in iterable:
        self[key] = value

  def __setitem__(self, key, value):
    assert isinstance(key, six.string_types)
    if not isinstance(value, ConfidenceReportEntry):
      raise TypeError("`value` must be a ConfidenceReportEntry, but got " + str(value) + " of type " + str(type(value)))
    super(ConfidenceReport, self).__setitem__(key, value)

class ConfidenceReportEntry(object):
  """
  A data structure reporting how much confidence a model assigned to its
  predictions on each example and whether those predictions were correct.

  :param correctness: ndarray, one bool per example indicating whether it was
    correct
  :param confidence: ndarray, one floating point value per example reporting
    the probability assigned to the prediction for that example
  """
  def __init__(self, correctness, confidence):
    assert isinstance(correctness, np.ndarray)
    assert isinstance(correctness, np.ndarray)
    assert correctness.ndim == 1
    assert confidence.ndim == 1
    assert correctness.dtype == np.bool
    assert np.issubdtype(confidence.dtype, np.floating)
    assert correctness.shape == confidence.shape
    assert confidence.min() >= 0.
    assert confidence.max() <= 1.
    self.correctness = correctness
    self.confidence = confidence

  def __getitem__(self, key):
    warnings.warn("Dictionary confidence report entries are deprecated."
                  "Switch to accessing the appropriate field of "
                  "ConfidenceReportEntry. "
                  "Dictionary-style access will be removed on or after "
                  "2019-04-24.")
    assert key in ['correctness', 'confidence']
    return self.__dict__[key]

  def __setitem__(self, key, value):
    warnings.warn("Dictionary confidence report entries are deprecated."
                  "Switch to accessing the appropriate field of "
                  "ConfidenceReportEntry. "
                  "Dictionary-style access will be removed on or after "
                  "2019-04-24.")
    assert key in ['correctness', 'confidence']
    self.__dict__[key] = value


def make_confidence_report_bundled(filepath, train_start=TRAIN_START,
                                   train_end=TRAIN_END, test_start=TEST_START,
                                   test_end=TEST_END, which_set=WHICH_SET,
                                   recipe=RECIPE, report_path=REPORT_PATH,
                                   nb_iter=NB_ITER, base_eps=None,
                                   base_eps_iter=None, base_eps_iter_small=None):
  """
  Load a saved model, gather its predictions, and save a confidence report.
  :param filepath: path to model to evaluate
  :param train_start: index of first training set example to use
  :param train_end: index of last training set example to use
  :param test_start: index of first test set example to use
  :param test_end: index of last test set example to use
  :param which_set: 'train' or 'test'
  :param nb_iter: int, number of iterations of attack algorithm
    (note that different recipes will use this differently,
     for example many will run two attacks, one with nb_iter
     iterations and one with 25X more)
  :param base_eps: float, epsilon parameter for threat model, on a scale of [0, 1].
    Inferred from the dataset if not specified.
  :param base_eps_iter: float, a step size used in different ways by different recipes.
    Typically the step size for a PGD attack.
    Inferred from the dataset if not specified.
  :param base_eps_iter_small: float, a second step size for a more fine-grained attack.
    Inferred from the dataset if not specified.
  """
  # Avoid circular import
  from cleverhans import attack_bundling
  if callable(recipe):
    run_recipe = recipe
  else:
    run_recipe = getattr(attack_bundling, recipe)

  # Set logging level to see debug information
  set_log_level(logging.INFO)

  # Create TF session
  sess = tf.Session()

  assert filepath.endswith('.joblib')
  if report_path is None:
    report_path = filepath[:-len('.joblib')] + "_bundled_report.joblib"

  with sess.as_default():
    model = load(filepath)
  assert len(model.get_params()) > 0
  factory = model.dataset_factory
  factory.kwargs['train_start'] = train_start
  factory.kwargs['train_end'] = train_end
  factory.kwargs['test_start'] = test_start
  factory.kwargs['test_end'] = test_end
  dataset = factory()

  center = dataset.kwargs['center']
  if 'max_val' in factory.kwargs:
    max_value = factory.kwargs['max_val']
  elif hasattr(dataset, 'max_val'):
    max_value = dataset.max_val
  else:
    raise AttributeError("Can't find max_value specification")
  min_value = 0. - center * max_value
  value_range = max_value - min_value

  if 'CIFAR' in str(factory.cls):
    if base_eps is None:
      base_eps = 8. / 255.
    if base_eps_iter is None:
      base_eps_iter = 2. / 255.
    if base_eps_iter_small is None:
      base_eps_iter_small = 1. / 255.
  elif 'MNIST' in str(factory.cls):
    if base_eps is None:
      base_eps = .3
    if base_eps_iter is None:
      base_eps_iter = .1
    base_eps_iter_small = None
  else:
    # Note that it is not required to specify base_eps_iter_small
    if base_eps is None or base_eps_iter is None:
      raise NotImplementedError("Not able to infer threat model from " + str(factory.cls))

  eps = base_eps * value_range
  eps_iter = base_eps_iter * value_range
  if base_eps_iter_small is None:
    eps_iter_small = None
  else:
    eps_iter_small = base_eps_iter_small * value_range
  clip_min = min_value
  clip_max = max_value

  x_data, y_data = dataset.get_set(which_set)
  assert x_data.max() <= max_value
  assert x_data.min() >= min_value

  assert eps_iter <= eps
  assert eps_iter_small is None or eps_iter_small <= eps

  # Different recipes take different arguments.
  # For now I don't have an idea for a beautiful unifying framework, so
  # we get an if statement.
  if recipe == 'random_search_max_confidence_recipe':
    # pylint always checks against the default recipe here
    # pylint: disable=no-value-for-parameter
    run_recipe(sess=sess, model=model, x=x_data, y=y_data, eps=eps,
               clip_min=clip_min, clip_max=clip_max, report_path=report_path)
  else:
    run_recipe(sess=sess, model=model, x=x_data, y=y_data,
               nb_classes=dataset.NB_CLASSES, eps=eps, clip_min=clip_min,
               clip_max=clip_max, eps_iter=eps_iter, nb_iter=nb_iter,
               report_path=report_path, eps_iter_small=eps_iter_small)


def print_stats(correctness, confidence, name):
  """
  Prints out accuracy, coverage, etc. statistics
  :param correctness: ndarray
    One bool per example specifying whether it was correctly classified
  :param confidence: ndarray
    The probability associated with each prediction
  :param name: str
    The name of this type of data (e.g. "clean", "MaxConfidence")
  """
  accuracy = correctness.mean()
  wrongness = 1 - correctness
  denom1 = np.maximum(1, wrongness.sum())
  ave_prob_on_mistake = (wrongness * confidence).sum() / denom1
  assert ave_prob_on_mistake <= 1., ave_prob_on_mistake
  denom2 = np.maximum(1, correctness.sum())
  ave_prob_on_correct = (correctness * confidence).sum() / denom2
  covered = confidence > 0.5
  cov_half = covered.mean()
  acc_half = (correctness * covered).sum() / np.maximum(1, covered.sum())
  print('Accuracy on %s examples: %0.4f' % (name, accuracy))
  print("Average prob on mistakes: %0.4f" % ave_prob_on_mistake)
  print("Average prob on correct: %0.4f" % ave_prob_on_correct)
  print("Accuracy when prob thresholded at .5: %0.4f" % acc_half)
  print("Coverage when prob thresholded at .5: %0.4f" % cov_half)

  success_rate = acc_half * cov_half
  # Success is correctly classifying a covered example
  print("Success rate at .5: %0.4f" % success_rate)
  # Failure is misclassifying a covered example
  failure_rate = (1. - acc_half) * cov_half
  print("Failure rate at .5: %0.4f" % failure_rate)
  print()


def make_confidence_report(filepath, train_start=TRAIN_START,
                           train_end=TRAIN_END,
                           test_start=TEST_START, test_end=TEST_END,
                           batch_size=BATCH_SIZE, which_set=WHICH_SET,
                           mc_batch_size=MC_BATCH_SIZE,
                           report_path=REPORT_PATH,
                           base_eps_iter=BASE_EPS_ITER,
                           nb_iter=NB_ITER, save_advx=SAVE_ADVX):
  """
  Load a saved model, gather its predictions, and save a confidence report.


  This function works by running a single MaxConfidence attack on each example.
  This provides a reasonable estimate of the true failure rate quickly, so
  long as the model does not suffer from gradient masking.
  However, this estimate is mostly intended for development work and not
  for publication. A more accurate estimate may be obtained by running
  make_confidence_report_bundled.py instead.

  :param filepath: path to model to evaluate
  :param train_start: index of first training set example to use
  :param train_end: index of last training set example to use
  :param test_start: index of first test set example to use
  :param test_end: index of last test set example to use
  :param batch_size: size of evaluation batches
  :param which_set: 'train' or 'test'
  :param mc_batch_size: batch size for MaxConfidence attack
  :param base_eps_iter: step size if the data were in [0,1]
    (Step size will be rescaled proportional to the actual data range)
  :param nb_iter: Number of iterations of PGD to run per class
  :param save_advx: bool. If True, saves the adversarial examples to disk.
    On by default, but can be turned off to save memory, etc.
  """

  # Set TF random seed to improve reproducibility
  tf.set_random_seed(1234)

  # Set logging level to see debug information
  set_log_level(logging.INFO)

  # Create TF session
  sess = tf.Session()

  if report_path is None:
    assert filepath.endswith('.joblib')
    report_path = filepath[:-len('.joblib')] + "_report.joblib"

  with sess.as_default():
    model = load(filepath)
  assert len(model.get_params()) > 0
  factory = model.dataset_factory
  factory.kwargs['train_start'] = train_start
  factory.kwargs['train_end'] = train_end
  factory.kwargs['test_start'] = test_start
  factory.kwargs['test_end'] = test_end
  dataset = factory()

  center = dataset.kwargs['center']
  max_val = dataset.kwargs['max_val']
  value_range = max_val * (1. + center)
  min_value = 0. - center * max_val

  if 'CIFAR' in str(factory.cls):
    base_eps = 8. / 255.
    if base_eps_iter is None:
      base_eps_iter = 2. / 255.
  elif 'MNIST' in str(factory.cls):
    base_eps = .3
    if base_eps_iter is None:
      base_eps_iter = .1
  else:
    raise NotImplementedError(str(factory.cls))

  mc_params = {'eps': base_eps * value_range,
               'eps_iter': base_eps_iter * value_range,
               'nb_iter': nb_iter,
               'clip_min': min_value,
               'clip_max': max_val}

  x_data, y_data = dataset.get_set(which_set)

  report = ConfidenceReport()

  semantic = Semantic(model, center, max_val, sess)
  mc = MaxConfidence(model, sess=sess)

  jobs = [('clean', None, None, None, False),
          ('Semantic', semantic, None, None, False),
          ('mc', mc, mc_params, mc_batch_size, True)]

  for job in jobs:
    name, attack, attack_params, job_batch_size, save_this_job = job
    if job_batch_size is None:
      job_batch_size = batch_size
    t1 = time.time()
    if save_advx and save_this_job:
      # If we want to save the adversarial examples to the filesystem, we need
      # to fetch all of them. Otherwise they're just computed one batch at a
      # time and discarded

      # The path to save to
      assert report_path.endswith('.joblib')
      advx_path = report_path[:-len('.joblib')] + '_advx_' + name + '.npy'

      # Fetch the adversarial examples
      x_data = run_attack(sess, model, x_data, y_data, attack, attack_params,
                          batch_size=job_batch_size, devices=devices)

      # Turn off the attack so `correctness_and_confidence` won't run it a
      # second time.
      attack = None
      attack_params = None

      # Save the adversarial examples
      np.save(advx_path, x_data)

    # Run correctness and confidence evaluation on adversarial examples
    packed = correctness_and_confidence(sess, model, x_data, y_data,
                                        batch_size=job_batch_size,
                                        devices=devices,
                                        attack=attack,
                                        attack_params=attack_params)
    t2 = time.time()
    print("Evaluation took", t2 - t1, "seconds")
    correctness, confidence = packed

    report[name] = ConfidenceReportEntry(correctness=correctness,
                                         confidence=confidence)

    print_stats(correctness, confidence, name)

  save(report_path, report)
